# Assistant d'installation
import asyncio
import json
import os
import time
import traceback
import warnings

import discord
from dotenv import load_dotenv
import gspread
import sqlalchemy
from oauth2client.service_account import ServiceAccountCredentials

from lgrez import __version__


def export(varname):
    """Returns "export varname={formated value}\\n" """
    var = str(globals()[varname])
    frmted_var = var.strip("'\"").replace("\"", "\\\"").replace(r"\n", r"\\n")
    return f"export {varname}=\"{frmted_var}\"\n"


def report_error(exc):
    print(f"Hmm, something went wrong ({type(exc).__name__}): {exc}")
    pft = input("Print the full traceback? (y/n) ")
    if pft.strip().lower() in ["y", "o", "yes"]:
        print(traceback.format_exc())

    input(
        "Check that given values are correct and that every steps\n"
        "mentionned above have been done, then press Enter to retry."
    )
    print("")


print(f"LG-Rez v{__version__} - Configuration Assistant Tool\n")

# Déjà en cours ?
print("Looking for existing installation in this folder...")

try:
    with open(".env", "r"):
        exists = True
except FileNotFoundError:
    exists = False

step = 0
if exists:
    print("Found! Checking progress...")

    load_dotenv(".env")

    if (LGREZ_DATABASE_URI := os.getenv("LGREZ_DATABASE_URI")):
        step = 1
        print("    step 1 ok")

    if (LGREZ_DISCORD_TOKEN := os.getenv("LGREZ_DISCORD_TOKEN")):
        step = 2
        print("    step 2 ok")
    if (LGREZ_SERVER_ID := os.getenv("LGREZ_SERVER_ID")):
        step = 3
        print("    step 3 ok")

    if (LGREZ_GCP_CREDENTIALS := os.getenv("LGREZ_GCP_CREDENTIALS")):
        step = 4
        print("    step 4 ok")
    if ((LGREZ_TDB_SHEET_ID := os.getenv("LGREZ_TDB_SHEET_ID"))
        and (LGREZ_ROLES_SHEET_ID := os.getenv("LGREZ_ROLES_SHEET_ID"))
        and (LGREZ_DATA_SHEET_ID := os.getenv("LGREZ_DATA_SHEET_ID"))):

        step = 5
        print("    step 5 ok")

    if (LGREZ_CONFIG_STATUS := os.getenv("LGREZ_CONFIG_STATUS")):
        step = 6
        print("    step 6 ok")

    if step == 6:
        print(
            "\nInstallation already complete in this folder.\n"
            "To create a fresh installation, delete the .env file;\n"
            "otherwise, just edit directly the variables it contains."
        )
        exit(0)

    input(
        f"Found ongoing installation on step {step + 1}/6. "
        "Press Enter to continue installation."
    )

else:
    print("Nothing found.")

# Base de .env
if step == 0:
    input("""
-----------------------------------------------------

Welcome to the LG-Rez Installation Assistant! Il will guide you through
the whole process of making a functionnal installation for your bot.

You can pause the installation anytime by killing this assistant, it
will resume at the current step.
Press Enter to begin.""")

    content = """
# .env: file containing text variables, considered by Python as environment
# variables with dotenv.loadenv() and then accessible through os.getenv.
# Here should be put:
#   - everything sensitive (acces tokens to different APIs, ...) ;
#   - every parameter that is season-dependant (server ID, ...).

# ALL VARIABLES WILL BE READ AS STRINGS, even if written without quote marks.
# Itegers/... stored here will need to be transformed after os.getenv call.

"""

    with open(".env", "w") as fich:
        fich.write(content.lstrip())

if step < 1:
    print("""\n------ STEP 1 : database connection ------

This program needs to be connected to a database.

Since it uses SQLAlchemy, every language it supports can theoretically
be used, but the package has been developed and tested with PostgreSQL
exclusively (otherwise, you may need to install complementary packages
such as PyMySQL).

You will need an empty database, local or on a specified host.
The database schema will be created by the bot the first time it runs.
""")

    ok = False
    while not ok:
        LGREZ_DATABASE_URI = input(
            "Database URI (dialect[+driver]://user:password@host/base): "
        )

        print("Testing connection...")
        try:
            engine = sqlalchemy.create_engine(LGREZ_DATABASE_URI)
            engine.connect()
        except Exception as e:
            report_error(e)
        else:
            print("Connected!")
            ok = True
            time.sleep(1)

    with open(".env", "a") as fich:
        fich.write("# -- Database\n\n" + export("LGREZ_DATABASE_URI"))
    step = 1


if step < 2:
    print("""\n\n------ STEP 2 : Discord application ------

If not already done, create a new application on the Discord
Developer Portal: https://discord.com/developers/applications
Name it as you wish (you will be able to rename the bot in your
server after) and click "Add bot" under the "Bot" section.

Here, turn on the two options under "Privilegied Gateway Intents".
You may want to uncheck "Public bot" too to have finest control.

Then, clic "Copy" under the "Token" session and paste it here.
This token gives anyone full access to your bot, so don't leak it!
""")

    ok = False
    while not ok:
        LGREZ_DISCORD_TOKEN = input("Discord client secret: ")

        print("Testing connection...")
        client = discord.Client()
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(client.login(LGREZ_DISCORD_TOKEN))
        except Exception as e:
            report_error(e)
        else:
            print("Connected!")
            loop.run_until_complete(client.logout())
            ok = True
            time.sleep(1)
        finally:
            loop.close()


    with open(".env", "a") as fich:
        fich.write("\n\n# -- Discord\n\n" + export("LGREZ_DISCORD_TOKEN"))
    step = 2


if step < 3:
    print("""\n\n------ STEP 3 : Discord server ------

For each season, we recommand you to create a new Discord server.
We provide the following model, with recommanded channels and permissions:
https://discord.new/RPGXChXXcKQZ
""")

    ok = False
    while not ok:
        LGREZ_SERVER_ID = input("Server ID (Server settings / Widget): ")

        try:
            if not(LGREZ_SERVER_ID.isdigit() and len(LGREZ_SERVER_ID) == 18):
                raise ValueError("Server ID must be a 18-digits number")
        except Exception as e:
            report_error(e)
        else:
            ok = True
            time.sleep(1)

    with open(".env", "a") as fich:
        fich.write(export("LGREZ_SERVER_ID"))

    input("""
Unfortunately, a Discord model doesn't contain the following:
 - Icon: create your own!
 - Emojis: the bot itself needs a few custom emojis to work correctly.
We provide them in a Google Drive "Templates" folder:
https://drive.google.com/drive/folders/1kjHzUSp-QfgI77Yg0GCxdM6YcFkHPYVw.

Download them, go in your server settings / Emoji / Upload emoji and
select them all; they well be named correctly.
When done, press Enter to continue.""")

    input("""Now, you can add the bot to your server:
  - Go to the Discord Developer Portal, select your bot application;
  - Go to the OAuth2 tab, select the "bot" scope and clic "Copy".
    Bot permissions are not important in our settings;
  - Open the copied URL in your browser, select the newly created server
    and clic "Authorize".
  - Go back to Discord and give the "Bot" role to the bot account.
    That will grant him every needed permissions.

The bot should appear offline: you'll be able to run it at the end of
this installation.
ONCE THE BOT IS SHOWN IN YOUR SERVER MEMBERS, press Enter to continue.""")

    step = 3


if step < 4:
    print("""\n\n------ STEP 4 : GCP connection ------

The bot needs a "special" Google account to access Google Sheets files.
If not already done, create a new projet on Google Cloud Platform :
https://console.cloud.google.com/home/dashboard
(you may encounter extra steps if you use GCP for the first time).

Follow the basic setup guide, then open the "IAM & Admin" panel,
go to "Service Accounts" and create a new service account.
Name it as you wish, then open it and create a new key ("Keys"/"Add key").
Open the JSON file, copy its contents and delete it.
""")

    ok = False
    while not ok:
        LGREZ_GCP_CREDENTIALS = input("JSON key (one-line file contents): ")

        print("Authentificating...")
        try:
            scope = ['https://spreadsheets.google.com/feeds']
            with warnings.catch_warnings(record=True) as warns:
                creds = ServiceAccountCredentials.from_json_keyfile_dict(
                    json.loads(LGREZ_GCP_CREDENTIALS),
                    scope
                )
                client = gspread.Client(creds)
                if warns:
                    raise RuntimeError(warns[0].message)

        except Exception as e:
            report_error(e)
        else:
            print("OK!")
            ok = True
            time.sleep(1)


    with open(".env", "a") as fich:
        fich.write(
            "\n\n# -- Google Sheets\n\n" + export("LGREZ_GCP_CREDENTIALS")
        )

    step = 4

if step < 5:
    print("""\n\n------ STEP 5 : Sheets IDs ------

This bot works with Google Sheets worksheets to gather players orders
and resolve play situations. We provide a readonly Google Drive folder
with every needed files:
https://drive.google.com/drive/folders/1kjHzUSp-QfgI77Yg0GCxdM6YcFkHPYVw

If not already done, duplicate all files into your own folder
(right clic > Create a copy) and open them. The bot needs the filed
IDs, ie the URL part between '/d/' and '/edit'.

(for a new season, we recommand you to create a new "Tableau de bord" file.
The faster way is to duplicate it again from our template folder, but you
may prefer duplicate one of your previous seasons and remove players and
sheets: in this case, remember to remove cached properties of players
(A-I columns).)

IMPORTANT: to allow your bot to read and write the sheets, share the three
(or the whole folder) with the service account mail adress created just
before (account@project.iam.gserviceaccount.com), with Editor rights.

NOTE: if your GCP project is newly created, you may first encourter an
error containing a link to visit to enable Google Sheets API for your
project.
""")

    ok = False
    while not ok:
        LGREZ_TDB_SHEET_ID = input("'Tableau de bord' ID: ")
        LGREZ_ROLES_SHEET_ID = input("'Roles et actions' ID: ")
        LGREZ_DATA_SHEET_ID = input("'Donnees brutes' ID: ")

        print("Checking access...")
        try:
            scope = ['https://spreadsheets.google.com/feeds']
            creds = ServiceAccountCredentials.from_json_keyfile_dict(
                json.loads(LGREZ_GCP_CREDENTIALS),
                scope
            )
            client = gspread.authorize(creds)

            workbook = client.open_by_key(LGREZ_TDB_SHEET_ID)
            workbook.sheet1.add_rows(1)     # Test en écriture
            print("'Tableau de bord' OK...")

            workbook = client.open_by_key(LGREZ_ROLES_SHEET_ID)
            workbook.sheet1.add_rows(1)     # Test en écriture
            print("'Roles et actions' OK...")

            workbook = client.open_by_key(LGREZ_DATA_SHEET_ID)
            workbook.sheet1.add_rows(1)     # Test en écriture

        except Exception as e:
            report_error(e)
        else:
            print("'Donnees brutes' OK!")
            ok = True
            time.sleep(1)

    with open(".env", "a") as fich:
        fich.write(
            "\n"
            + export("LGREZ_TDB_SHEET_ID")
            + export("LGREZ_ROLES_SHEET_ID")
            + export("LGREZ_DATA_SHEET_ID")
            + """
# Reminder: the sheets need to be shared with GCP_CREDENTIALS["client_email"]
# (as an Editor) to grant the bot read and write rights.\n\n"""
        )
    step = 5

if step < 6:
    input("""\n\n------ STEP 6 : dashboard setup ------

In the "Tableau de bord" sheet, under "Journée en cours" worksheet, edit
the AN1 and BH1 cells to put the full URLs of the "Données brutes" and
"Rôles et actions" sheets, respectively.

Once done, "#REF!" should appear in AH3 and BH4 cells (among other).
A tooltip should appear when clicking on them, asking you to grant
access to those files: do it.

Press Enter when done.
""")

    print("""Now, open the scripts editor ("Tools/Scripts editor"), select
the "Triggers" tab on the left (alarm clock).
This tab will allow us to automatically backup and clear the dashboard
each day: clic "Add a trigger" and configure it like (left, up to down):
Backupfeuille / Head / Time trigger / Daily / Between 1am and 2am
(important, because the data sheet clears itself between 3pm and 4pm).

You may want to do that later, as it will create a new worksheet each
day (except weekends) while enabled.

Press Enter when done.""")
    input()

    LGREZ_CONFIG_STATUS = 1
    with open(".env", "a") as fich:
        fich.write(
            "# -- Configuration status: set it to \"1\" when everything "
            "is functionnal\n\n"
            + export("LGREZ_CONFIG_STATUS")
        )        # Fin de l'installation
    step = 6


# Création de start_bot.py
code = """# Minimal working code to run LG-bot!

from lgrez import LGBot

bot = LGBot()
bot.run()
"""

with open("start_bot.py", "w") as fich:
    fich.write(code)


print("""\n------ THE END ------

Congrats, the installation is now complete! Variables have been
written to the ".env" file: edit it directly for minor changes.
A file named "start_bot.py" has been created in the current folder:
it contains the minimal code needed to run the bot. It will be running
unless it crashes or is manually killed.

The bot execution is quiet, except at startup and if an exception
occurs, so we advise you to log the output stream somewhere!
We also advise you to use an external script to ensure the bot has
not crashed -- pretty unlikely, but it might always happen; see also
"output_liveness" customization option in the docs:
https://lg-rez.readthedocs.io/fr/2.0.0/config.html#lgrez.config.output_liveness

You can no try to interact with the bot!
(For your first tests, note that the bot ignore every messages posted by
a member without any role! Assign yourself the "MJ" role to get full
rights with the bot.)

Do not hesitate to reach us for any issue or suggestion, we're always
happy to get news. See bottom of README for contacts.

Enjoy!""")



import unittest
from unittest import mock

import discord

from lgrez import config, bdd, blocs, features, __version__
from lgrez.features import special
from test import mock_discord



class TestSpecialFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.special functions."""

    @mock.patch("lgrez.blocs.one_command.bypass")
    async def test__filter_runnables(self, bp_patch):
        """Unit tests for _filter_runnables function."""
        # async def _filter_runnables(commands, ctx)
        _filter_runnables = special._filter_runnables

        commands = [
            mock.Mock(can_run=mock.AsyncMock(return_value=True)),
            mock.Mock(can_run=mock.AsyncMock(return_value=False)),
            mock.Mock(can_run=mock.AsyncMock(side_effect=RuntimeError)),
        ]
        ctx = mock.Mock()
        res = await _filter_runnables(commands, ctx)
        bp_patch.assert_called_once_with(ctx)
        bp_patch.return_value.__enter__.assert_called_once()
        self.assertEqual(res, [commands[0]])



class TestSpecial(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.special commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = special.Special(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()


    async def test_panik(self):
        """Unit tests for !panik command."""
        # async def panik(self, ctx)
        panik = self.cog.panik

        ctx = mock_discord.get_ctx(panik)
        with mock.patch("sys.exit") as exit_patch:
            await ctx.invoke()
        exit_patch.assert_called_once()


    async def test_do(self):
        """Unit tests for !do command."""
        # async def do(self, ctx, *, code)
        do = self.cog.do

        samples = {
            "1": 1,
            "1 + 2": 3,
            "37*(5-2)": 37*(5-2),
            "config.debut_saison": config.debut_saison,
            "blocs.tools.montre()": blocs.tools.montre(),
            "tools.montre()": blocs.tools.montre(),
            "features.sync.transtype": features.sync.transtype,
            "bdd.Joueur": bdd.Joueur,
            "Joueur": bdd.Joueur,
        }

        # Vérification bases + accès à tous modules
        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(do, code=sample)
            await ctx.invoke()
            ctx.assert_sent(str(result))

        # Vérification coroutines
        ctx = mock_discord.get_ctx(do, code="ctx.send('oh' + 'ah')")
        ctx.send.return_value = "bookz"
        await ctx.invoke()
        ctx.assert_sent("ohah", "bookz")

        # Vérification erreurs
        ctx = mock_discord.get_ctx(do, code="blbllb")
        with self.assertRaises(NameError):
            await ctx.invoke()
        ctx.assert_sent()


    @mock.patch("lgrez.blocs.realshell.RealShell")
    async def test_shell(self, rs_patch):
        """Unit tests for !shell command."""
        # async def shell(self, ctx)
        shell = self.cog.shell
        rs_patch().interact = mock.AsyncMock()
        rs_patch.reset_mock()

        # Okay
        ctx = mock_discord.get_ctx(shell)
        await ctx.invoke()
        rs_patch.assert_called_once_with(ctx.channel, mock.ANY)
        self.assertEqual(rs_patch.call_args.args[1]["ctx"], ctx)
        ctx.assert_sent()
        rs_patch.reset_mock()

        # Exit
        rs_patch().interact.side_effect = blocs.realshell.RealShellExit
        rs_patch.reset_mock()
        ctx = mock_discord.get_ctx(shell)
        with self.assertRaises(blocs.tools.CommandExit):
            await ctx.invoke()
        rs_patch.assert_called_once_with(ctx.channel, mock.ANY)
        self.assertEqual(rs_patch.call_args.args[1]["ctx"], ctx)
        ctx.assert_sent()


    @mock.patch("lgrez.blocs.tools.member")
    @mock.patch("lgrez.features.inscription.main")
    async def test_co(self, im_patch, mb_patch):
        """Unit tests for !co command."""
        # async def co(self, ctx, cible=None)
        co = self.cog.co

        # no cible
        ctx = mock_discord.get_ctx(co)
        await ctx.invoke()
        im_patch.assert_called_once_with(ctx.author)
        mb_patch.assert_not_called()
        im_patch.reset_mock()

        # cible = non-existing player
        mb_patch.side_effect = ValueError
        ctx = mock_discord.get_ctx(co, cible="booo")
        await ctx.invoke()
        ctx.assert_sent("introuvable")
        im_patch.assert_not_called()
        mb_patch.reset_mock(side_effect=True)

        # cible = existing player
        ctx = mock_discord.get_ctx(co, cible="noice")
        await ctx.invoke()
        mb_patch.assert_called_once_with("noice")
        im_patch.assert_called_once_with(mb_patch.return_value)


    @mock.patch("lgrez.blocs.tools.boucle_query_joueur")
    @mock.patch("lgrez.blocs.one_command.bypass")
    async def test_doas(self, ocb_patch, bqj_patch):
        """Unit tests for !doas command."""
        # async def doas(self, ctx, *, qui_quoi)
        doas = self.cog.doas

        # bad argument
        ctx = mock_discord.get_ctx(doas, qui_quoi="brbbrbr")
        with self.assertRaises(discord.ext.commands.UserInputError):
            await ctx.invoke()
        ctx = mock_discord.get_ctx(doas, qui_quoi="brbbr!br")
        with self.assertRaises(discord.ext.commands.UserInputError):
            await ctx.invoke()
        ocb_patch.assert_not_called()
        bqj_patch.assert_not_called()
        config.bot.process_commands.assert_not_called()

        # proceed
        ctx = mock_discord.get_ctx(doas, qui_quoi="bizz !kmd argz")
        await ctx.invoke()
        bqj_patch.assert_called_once_with(ctx, "bizz")
        ocb_patch.assert_called_once_with(ctx)
        config.bot.process_commands.assert_called_once_with(ctx.message)
        self.assertEqual(ctx.message.author, bqj_patch.return_value.member)
        self.assertEqual(ctx.message.content, "!kmd argz")


    @mock.patch("lgrez.blocs.one_command.bypass")
    async def test_secret(self, ocb_patch):
        """Unit tests for !secret command."""
        # async def secret(self, ctx, *, quoi)
        secret = self.cog.secret

        # proceed
        ctx = mock_discord.get_ctx(secret, quoi="krkrkrk")
        await ctx.invoke()
        ctx.message.delete.assert_called_once()
        ocb_patch.assert_called_once_with(ctx)
        config.bot.process_commands.assert_called_once_with(ctx.message)
        self.assertEqual(ctx.message.content, "krkrkrk")


    async def test_stop(self):
        """Unit tests for !stop command."""
        # async def stop(self, ctx)
        stop = self.cog.stop
        config.bot.in_command = [11, 12]

        # not in bot.in_command
        ctx = mock_discord.get_ctx(stop)
        ctx.channel.id = 13
        await ctx.invoke()
        self.assertEqual(config.bot.in_command, [11, 12])

        # in bot.in_command
        ctx = mock_discord.get_ctx(stop)
        ctx.channel.id = 12
        await ctx.invoke()
        self.assertEqual(config.bot.in_command, [11])


    async def test_help(self):
        """Unit tests for !help command."""
        # async def help(self, ctx, *, command=None)
        help = self.cog.help

        def _mock_cmd(name, help, short_doc, signature, cog=None,
                      runnable=True, aliases=[]):
            cmd = mock.MagicMock(discord.ext.commands.Command, name=name)
            cmd.configure_mock(
                name=name,
                help=help,
                short_doc=short_doc,
                signature=signature,
                cog=cog,
                can_run=mock.AsyncMock(return_value=runnable),
                aliases=aliases,
            )
            return cmd

        def _mock_cog(name, description, commands=[]):
            CogMock = type(name, (mock.MagicMock, ), {})
            cog = CogMock(discord.ext.commands.Cog, name=name)
            cog.configure_mock(
                description=description,
                commands=commands,
            )
            return cog

        compdoc = """Shorthand line.

        Descriptif de la commande, a ``texte Sphinx brut`` b
        Args:
            oui les arguments
        Warning:
            attention !

        :objet:`énorme`: klakla, :allo: !
        """

        cogs = {
            "cog1": _mock_cog("Cog1", "cogzz1"),
            "cog2": _mock_cog("Cog2", "cogzz2"),
            "cog3": _mock_cog("Cog3", "cogzz3"),
            "cog4": _mock_cog("Cog4", "cogzz4"),
            }
        cmds = [
            _mock_cmd("cmd1", "commz1\n\ndoc1", "szdk1", "sig1",
                      runnable=True, cog=cogs["cog1"]),
            _mock_cmd("cmd2", "commz2\n\ndoc2", "szdk2", "sig2",
                      runnable=False, cog=cogs["cog1"], aliases=["okm2"]),
            _mock_cmd("cm2b", "comm2b\n\ndo2b", "szd2b", "si2b",
                      runnable=True, cog=cogs["cog1"], aliases=["okm2b"]),
            _mock_cmd("cmd3", "commz3\n\ndoc3", "szdk3", "sig3",
                      runnable=True, cog=cogs["cog2"],
                      aliases=["okm31", "okm32"]),
            _mock_cmd("cmd4", f"commz4\n\n{compdoc}", "szdk4", "sig4",
                      runnable=True),
            _mock_cmd("cmd5", "commz5\n\ndoc5", "szdk5", "sig5",
                      runnable=False, cog=cogs["cog3"], aliases=["okm5"]),
        ]
        cogs["cog1"].get_commands.return_value = cmds[:3]
        cogs["cog2"].get_commands.return_value = [cmds[3]]
        cogs["cog3"].get_commands.return_value = [cmds[5]]

        config.bot.description = "deskrzz"
        config.bot.commands = cmds
        config.bot.cogs = cogs

        # list commands
        ctx = mock_discord.get_ctx(help)
        await ctx.invoke()
        ctx.assert_sent([
            "deskrzz", f"v{__version__}",
            "Cog1", "cogzz1", "!cmd1", "szdk1", "!cm2b", "szd2b",
            "Cog2", "cogzz2", "!cmd3", "szdk3",
            "Commandes isolées", "!cmd4", "szdk4",
        ])
        ctx.assert_not_sent([
            "Cog3", "cogzz3", "Cog4", "cogzz4",
            "!cmd2", "szdk2", "!cmd5", "szdk5",
        ])

        # !help <unknown command>
        ctx = mock_discord.get_ctx(help, command="blebele")
        await ctx.invoke()
        ctx.assert_sent("Commande '!blebele' non trouvée")

        # !help cmd1
        ctx = mock_discord.get_ctx(help, command="cmd1")
        await ctx.invoke()
        ctx.assert_sent([
            "!cmd1", "sig1", "commz1", "doc1"
        ])

        # !help !cmd2
        ctx = mock_discord.get_ctx(help, command="!cmd2")
        await ctx.invoke()
        ctx.assert_sent([
            "!cmd2", "sig2", "commz2", "doc2", "okm2"
        ])

        # !help ! cmd2b
        ctx = mock_discord.get_ctx(help, command="! cm2b")
        await ctx.invoke()
        ctx.assert_sent([
            "!cm2b", "si2b", "comm2b", "do2b", "okm2b"
        ])

        # !help okm31 (alias)
        ctx = mock_discord.get_ctx(help, command="okm31")
        await ctx.invoke()
        ctx.assert_sent([
            "!cmd3", "sig3", "commz3", "doc3", "okm31", "okm32"
        ])

        # !help cmd4 (complex doc, formatted)
        ctx = mock_discord.get_ctx(help, command="cmd4")
        await ctx.invoke()
        ctx.assert_sent([
            "!cmd4", "sig4", "commz4", "Shorthand line.",
            "a `texte Sphinx brut` b", "Arguments :", "Avertissement :",
            "`énorme`: klakla", ":allo:",
        ])
        ctx.assert_not_sent([
            "``texte", "Args:", "Warning:", ":objet:",
        ])

        # !help {otherprefix} okm5 (alias)
        config.bot.command_prefix = "@<>"
        ctx = mock_discord.get_ctx(help, command="@<> okm5")
        await ctx.invoke()
        ctx.assert_sent([
            "@<>cmd5", "sig5", "commz5", "doc5", "okm5"
        ])



    async def test_apropos(self):
        """Unit tests for !apropos command."""
        # async def apropos(self, ctx)
        apropos = self.cog.apropos
        config.bot.description = "deskrzz"
        config.bot.user.avatar_url = "uzavurlz"

        ctx = mock_discord.get_ctx(apropos)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("embed", ctx.send.call_args.kwargs)
        embed = ctx.send.call_args.kwargs["embed"]
        self.assertIn(f"v{__version__}", embed.title)
        self.assertEqual("deskrzz", embed.description)
        self.assertEqual("uzavurlz", embed.author.icon_url)
        self.assertIn("logo_espci.png", embed.image.url)



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

import sqlalchemy

from lgrez import config
from lgrez.bdd import base
from lgrez.bdd.base import (autodoc_Column, autodoc_ManyToOne,
                            autodoc_OneToMany)


# Tables de données

class Reaction(base.TableBase):
    """Table de données des réactions d'IA connues du bot.

    Les instances sont enregistrées via :meth:`\!addIA
    <.IA.GestionIA.GestionIA.addIA.callback>` et supprimées via
    :meth:`\!modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique de la réaction, sans signification")
    reponse = autodoc_Column(sqlalchemy.String(2000), nullable=False,
        doc="Réponse, suivant le format (mini-langage) personnalisé "
            "(``\"txt <||> txt <&&> <##>react\"``)")

    # One-to-manys
    triggers = autodoc_OneToMany("Trigger", back_populates="reaction",
        doc="Déclencheurs associés")

    def __repr__(self):
        """Return repr(self)."""
        extract = self.reponse.replace('\n', ' ')[:15] + "..."
        return f"<Reaction #{self.id} ({extract})>"


class Trigger(base.TableBase):
    """Table de données des mots et expressions déclenchant l'IA du bot.

    Les instances sont enregistrées via :meth:`\!addIA
    <.IA.GestionIA.GestionIA.addIA.callback>` ou :meth:`\!modifIA
    <.IA.GestionIA.GestionIA.modifIA.callback>` et supprimées via
    :meth:`\!modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique du déclencheur, sans signification")
    trigger = autodoc_Column(sqlalchemy.String(500), nullable=False,
        doc="Mots-clés / expressions")

    _reac_id = sqlalchemy.Column(sqlalchemy.ForeignKey("reactions.id"),
        nullable=False)
    reaction = autodoc_ManyToOne("Reaction", back_populates="triggers",
        doc="Réaction associée")

    def __repr__(self):
        """Return repr(self)."""
        extract = self.trigger.replace('\n', ' ')[:15] + "..."
        return f"<Trigger #{self.id} ({extract})>"



"""lg-rez / features / Commandes annexes

Commandes diverses qu'on ne savait pas où ranger

"""

import random
import requests
import datetime

from discord.ext import commands
from akinator.async_aki import Akinator

from lgrez.blocs import tools
from lgrez.bdd import Joueur, Role, Camp


class Annexe(commands.Cog):
    """Commandes annexes aux usages divers"""

    @commands.command()
    async def roll(self, ctx, *, XdY):
        """Lance un ou plusieurs dés

        Args:
            XdY: dés à lancer + modifieurs, au format
                ``XdY + XdY + ... + Z - Z ...`` avec X le nombre de dés,
                Y le nombre de faces et Z les modifieurs (constants) ;
                OU roll spécial : ``joueur`` / ``vivant`` / ``mort`` /
                ``rôle`` / ``camp``.

        Examples:
            - ``!roll 1d6``           -> lance un dé à 6 faces
            - ``!roll 1d20 +3``       -> lance un dé à 20 faces,
              ajoute 3 au résultat
            - ``!roll 1d20 + 2d6 -8`` -> lance un dé 20 plus deux dés 6,
              enlève 8 au résultat
            - ``!roll vivant``        -> choisit un joueur vivant
        """
        # Rolls spéciaux
        if XdY.lower() in ["joueur", "joueurs"]:
            await ctx.reply(random.choice(Joueur.query.all()).nom)
            return
        if XdY.lower() in ["vivant", "vivants"]:
            await ctx.reply(random.choice(
                Joueur.query.filter(Joueur.est_vivant).all()
            ).nom)
            return
        if XdY.lower() in ["mort", "morts"]:
            await ctx.reply(random.choice(
                Joueur.query.filter(Joueur.est_mort).all()
            ).nom)
            return
        if XdY.lower() in ["role", "rôle", "roles", "rôles"]:
            await ctx.reply(random.choice(
                Role.query.filter_by(actif=True).all()
            ).nom_complet)
            return
        if XdY.lower() in ["camp", "camps"]:
            await ctx.reply(random.choice(
                Camp.query.filter_by(public=True).all()
            ).nom)
            return

        dices = XdY.replace(' ', '').replace('-', '+-').split('+')
        # "1d6 + 5 - 2" -> ["1d6", "5", "-2"]
        r = ""
        s = 0
        try:
            for dice in dices:
                if 'd' in dice:
                    nb, faces = dice.split('d', maxsplit=1)
                    for i in range(int(nb)):
                        v = random.randrange(int(faces)) + 1
                        s += v
                        r += f" + {v}₍{tools.sub_chiffre(int(faces), True)}₎"
                else:
                    v = int(dice)
                    s += v
                    r += f" {'-' if v < 0 else '+'} {abs(v)}"
            r += f" = {tools.emoji_chiffre(s, True)}"
        except Exception:
            raise commands.UserInputError(f"pattern non reconu")
        else:
            await tools.send_blocs(ctx, r[3:])


    @commands.command(aliases=["cf", "pf"])
    async def coinflip(self, ctx):
        """Renvoie le résultat d'un tirage à Pile ou Face (aléatoire)

        Pile je gagne, face tu perds.
        """
        await ctx.send(random.choice(["Pile", "Face"]))


    @commands.command(aliases=["pong"])
    async def ping(self, ctx):
        """Envoie un ping au bot

        Pong

        Warning:
            Commande en bêta, non couverte par les tests unitaires.
        """
        ts_rec = datetime.datetime.utcnow()
        delta_rec = ts_rec - ctx.message.created_at
        # Temps de réception = temps entre création message et sa réception
        pingpong = ctx.invoked_with.replace("i", "x").replace("I", "X")
        pingpong = pingpong.replace("o", "i").replace("O", "I")
        pingpong = pingpong.replace("x", "o").replace("X", "O")

        cont = (
            f" Réception : {delta_rec.total_seconds()*1000:4.0f} ms\n"
            f" Latence :   {ctx.bot.latency*1000:4.0f} ms\n"
        )
        mess = await ctx.send(
            f"!{pingpong}\n" + tools.code_bloc(cont + " (...)")
        )

        ts_ret = datetime.datetime.utcnow()
        delta_ret = ts_ret - mess.created_at
        # Retour information message réponse créé
        delta_env = ts_ret - ts_rec - delta_ret
        # Temps d'envoi = temps entre réception 1er message (traitement quasi
        # instantané) et création 2e, moins temps de retour d'information
        delta_tot = delta_rec + delta_ret + delta_env
        # Total = temps entre création message !pong et réception information
        # réponse envoyée
        await mess.edit(content=f"!{pingpong}\n" + tools.code_bloc(
            cont +
            f" Envoi :     {delta_env.total_seconds()*1000:4.0f} ms\n"
            f" Retour :    {delta_ret.total_seconds()*1000:4.0f} ms\n"
            f"——————————————————————\n"
            f" Total :     {delta_tot.total_seconds()*1000:4.0f} ms"
        ))


    @commands.command()
    async def akinator(self, ctx):
        """J'ai glissé chef

        Implémentation directe de https://pypi.org/project/akinator.py

        Warning:
            Commande en bêta, non couverte par les tests unitaires.
        """
        # Un jour mettre ça dans des embeds avec les https://fr.akinator.com/
        # bundles/elokencesite/images/akitudes_670x1096/<akitude>.png croppées,
        # <akitude> in ["defi", "serein", "inspiration_legere",
        # "inspiration_forte", "confiant", "mobile", "leger_decouragement",
        # "vrai_decouragement", "deception", "triomphe"]
        await ctx.send(
            "Vous avez demandé à être mis en relation avec "
            + tools.ital("Akinator : Le Génie du web")
            + ".\nVeuillez patienter..."
        )
        async with ctx.typing():
            # Connexion
            aki = Akinator()
            question = await aki.start_game(language="fr")

        exit = False
        while not exit and aki.progression <= 80:
            mess = await ctx.send(f"({aki.step + 1}) {question}")
            reponse = await tools.wait_for_react_clic(
                mess,
                {"👍": "yes", "🤷": "idk", "👎": "no", "⏭️": "stop"}
            )
            if reponse == "stop":
                exit = True
            else:
                async with ctx.typing():
                    question = await aki.answer(reponse)

        async with ctx.typing():
            await aki.win()

        mess = await ctx.send(
            f"Tu penses à {tools.bold(aki.first_guess['name'])} "
            f"({tools.ital(aki.first_guess['description'])}) !\n"
            f"J'ai bon ?\n{aki.first_guess['absolute_picture_path']}"
        )
        if await tools.yes_no(mess):
            await ctx.send(
                "Yay\nhttps://fr.akinator.com/bundles/elokencesite"
                "/images/akitudes_670x1096/triomphe.png"
            )
        else:
            await ctx.send(
                "Oof\nhttps://fr.akinator.com/bundles/elokencesite"
                "/images/akitudes_670x1096/deception.png"
            )


    @commands.command()
    async def xkcd(self, ctx, N):
        """J'ai aussi glissé chef, mais un peu moins

        Args:
            N: numéro du comic

        Warning:
            Commande en bêta, non couverte par les tests unitaires.
        """
        async with ctx.typing():
            r = requests.get(f"https://xkcd.com/{N}/info.0.json")

        if not r:
            await ctx.send("Paramètre incorrect ou service non accessible.")
            return

        url = r.json().get("img")
        if not url:
            await ctx.send("Paramètre incorrect ou service non accessible.")
            return

        await ctx.send(url)



"""lg-rez / features / Actions publiques

Gestion des haros, candidatures à la mairie, résultats des votes

"""

import discord
from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.bdd import Joueur, CandidHaro, Statut, CandidHaroType, Vote


class ActionsPubliques(commands.Cog):
    """Commandes d'actions vous engageant publiquement"""

    @commands.command()
    @tools.vivants_only
    @tools.private
    async def haro(self, ctx, *, cible=None):
        """Lance publiquement un haro contre un autre joueur.

        Args:
            cible: nom du joueur à accuser

        Cette commande n'est utilisable que lorsqu'un vote pour le
        condamné est en cours.
        """
        auteur = ctx.author
        joueur = Joueur.from_member(auteur)
        try:
            vaction = joueur.action_vote(Vote.cond)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        if not vaction.is_open:
            await ctx.send("Pas de vote pour le condamné du jour en cours !")
            return

        cible = await tools.boucle_query_joueur(
            ctx, cible,
            "Contre qui souhaites-tu déverser ta haine ?"
        )

        if cible.statut == Statut.mort:
            await ctx.send("Nan mais oh, tu crois qu'il a pas assez "
                           "souffert en mourant lui ?")
            return

        elif cible.statut == Statut.immortel:
            await ctx.send("Comment oses-tu t'en prendre à ceux qui te sont "
                           f"supérieurs ? {config.Role.mj.mention}, regardez "
                           "un peu ce qu'il se passe là...")
            return

        await tools.send_blocs(
            ctx,
            "Et quelle est la raison de cette haine, d'ailleurs ?"
        )
        motif = await tools.wait_for_message_here(ctx)

        emb = discord.Embed(
            title=(f"**{config.Emoji.ha}{config.Emoji.ro} "
                   f"contre {cible.nom} !**"),
            description=f"**« {motif.content} »\n**",
            color=0xff0000
        )
        emb.set_author(name=f"{joueur.nom} en a gros 😡😡")
        emb.set_thumbnail(url=config.Emoji.bucher.url)
        emb.set_footer(
            text=f"Utilise !vote {cible.nom} pour voter contre cette personne."
        )

        mess = await ctx.send("C'est tout bon ?", embed=emb)
        if await tools.yes_no(mess):
            if not CandidHaro.query.filter_by(joueur=cible,
                                              type=CandidHaroType.haro).all():
                # Inscription haroté
                config.session.add(CandidHaro(joueur=cible,
                                              type=CandidHaroType.haro))

            if not CandidHaro.query.filter_by(joueur=joueur,
                                              type=CandidHaroType.haro).all():
                # Inscription haroteur
                config.session.add(CandidHaro(joueur=joueur,
                                              type=CandidHaroType.haro))

            config.session.commit()

            await config.Channel.haros.send(
                f"(Psst, {cible.member.mention} :3)",
                embed=emb
            )
            await config.Channel.debats.send(
                f"{config.Emoji.ha}{config.Emoji.ro} de {auteur.mention} "
                f"sur {cible.member.mention} ! Vous en pensez quoi vous ? "
                f"(détails sur {config.Channel.haros.mention})"
            )
            await ctx.send(
                f"Allez, c'est parti ! ({config.Channel.haros.mention})"
            )

        else:
            await ctx.send("Mission aborted.")


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def candid(self, ctx):
        """Candidate à l'élection du nouveau maire.

        Cette commande n'est utilisable que lorsqu'un vote pour le
        nouveau maire est en cours.
        """
        auteur = ctx.author
        joueur = Joueur.from_member(auteur)
        try:
            vaction = joueur.action_vote(Vote.maire)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        if not vaction.is_open:
            await ctx.send("Pas de vote pour le nouveau maire en cours !")
            return

        if CandidHaro.query.filter_by(joueur=joueur,
                                      type=CandidHaroType.candidature).all():
            await ctx.send(
                "Hola collègue, tout doux, tu t'es déjà présenté(e) !"
            )
            return

        await tools.send_blocs(ctx, "Quel est ton programme politique ?")
        motif = await tools.wait_for_message_here(ctx)

        emb = discord.Embed(
            title=(f"**{config.Emoji.maire} {joueur.nom} "
                   "candidate à la Mairie !**"),
            description=("Voici son programme politique :\n"
                         + tools.bold(motif.content)),
            color=0xf1c40f
        )
        emb.set_author(name=f"{joueur.nom} vous a compris !")
        emb.set_thumbnail(url=config.Emoji.maire.url)
        emb.set_footer(
            text=(f"Utilise !votemaire {auteur.display_name} "
                  "pour voter pour cette personne.")
        )

        mess = await ctx.send("C'est tout bon ?", embed=emb)
        if await tools.yes_no(mess):
            ch = CandidHaro(joueur=joueur, type=CandidHaroType.candidature)
            config.session.add(ch)
            config.session.commit()

            await config.Channel.haros.send(
                "Here comes a new challenger !",
                embed=emb
            )
            await config.Channel.debats.send(
                f"{auteur.mention} se présente à la Mairie ! "
                "Vous en pensez quoi vous ?\n"
                f"(détails sur {config.Channel.haros.mention})"
            )
            await ctx.send(
                f"Allez, c'est parti ! ({config.Channel.haros.mention})"
            )

        else:
            await ctx.send("Mission aborted.")


    @commands.command()
    @tools.mjs_only
    async def wipe(self, ctx, quoi):
        """Efface les haros / candidatures du jour (COMMANDE MJ)

        Args:
            quoi (str): peut être

                - ``haros`` : Supprimer les haros
                - ``candids`` : Supprimer les candicatures

        (commande non testée unitairement)
        """
        if quoi == "haros":
            cht = CandidHaroType.haro
        elif quoi == "candids":
            cht = CandidHaroType.candidature
        else:
            await ctx.send("Mauvais argument")

        items = CandidHaro.query.filter_by(type=cht).all()

        if not items:
            await ctx.send("Rien à faire")
            await tools.log(f"!wipe {quoi} : rien à faire")
        else:
            CandidHaro.delete(*items)
            await ctx.send("Fait.")
            await tools.log(f"!wipe {quoi} : fait")



"""lg-rez / features / Commandes spéciales

Commandes spéciales (méta-commandes, imitant ou impactant le
déroulement des autres ou le fonctionnement du bot)

"""

import asyncio
import re
import sys

# Unused imports because useful for !do / !shell globals
import discord
from discord.ext import commands

from lgrez import __version__, config, features, blocs, bdd
from lgrez.blocs import tools, realshell, one_command
from lgrez.bdd import *       # toutes les tables dans globals()


async def _filter_runnables(commands, ctx):
    """Retourne les commandes pouvant run parmis commands"""
    runnables = []
    with one_command.bypass(ctx):
        # On désactive la limitation de une commande simultanée
        # sinon can_run renvoie toujours False
        for cmd in commands:
            try:
                runnable = await cmd.can_run(ctx)
            except Exception:
                runnable = False
            if runnable:
                runnables.append(cmd)
    return runnables


class Special(commands.Cog):
    """Commandes spéciales (méta-commandes et expérimentations)"""

    @one_command.do_not_limit
    @commands.command(aliases=["kill"])
    @tools.mjs_only
    async def panik(self, ctx):
        """Tue instantanément le bot, sans confirmation (COMMANDE MJ)

        PAAAAANIK
        """
        sys.exit()


    @commands.command()
    @tools.mjs_only
    async def do(self, ctx, *, code):
        """Exécute du code Python et affiche le résultat (COMMANDE MJ)

        Args:
            code: instructions valides dans le contexte du LGbot
                (utilisables notemment : ``ctx``, ``config``, ``blocs``,
                ``features``, ``bdd``, ``<table>``...)

        Si ``code`` est une coroutine, elle sera awaited
        (ne pas inclure ``await`` dans ``code``).

        Aussi connue sous le nom de « faille de sécurité », cette
        commande permet de faire environ tout ce qu'on veut sur le bot
        (y compris le crasher, importer des modules, exécuter des
        fichiers .py... même si c'est un peu compliqué) voire d'impacter
        le serveur sur lequel le bot tourne si on est motivé.

        À utiliser avec parcimonie donc, et QUE pour du
        développement/debug !
        """
        class Answer:
            rep = None
        _a = Answer()

        locs = globals()
        locs["ctx"] = ctx
        locs["_a"] = _a
        exec(f"_a.rep = {code}", locs)
        if asyncio.iscoroutine(_a.rep):
            _a.rep = await _a.rep
        await tools.send_code_blocs(ctx, str(_a.rep))


    @commands.command()
    @tools.mjs_only
    async def shell(self, ctx):
        """Lance un terminal Python directement dans Discord (COMMANDE MJ)

        Envoyer ``help`` dans le pseudo-terminal pour plus
        d'informations sur son fonctionnement.

        Évidemment, les avertissements dans ``!do`` s'appliquent ici :
        ne pas faire n'imp avec cette commande !! (même si ça peut être
        très utile, genre pour ajouter des gens en masse à un channel)
        """
        locs = globals()
        locs["ctx"] = ctx
        shell = realshell.RealShell(ctx.channel, locs)
        try:
            await shell.interact()
        except realshell.RealShellExit as exc:
            raise tools.CommandExit(*exc.args or ["!shell: Forced to end."])


    @commands.command()
    @tools.mjs_only
    async def co(self, ctx, cible=None):
        """Lance la procédure d'inscription pour un membre (COMMANDE MJ)

        Fat comme si on se connectait au serveur pour la première fois.

        Args:
            cible: le nom exact ou la mention (``@joueur``) du joueur
                à inscrire, par défaut le lançeur de la commande.

        Cette commande est principalement destinée aux tests de
        développement, mais peut être utile si un joueur chibre son
        inscription (à utiliser dans son channel, ou ``#bienvenue``
        (avec ``!autodestruct``) si même le début a chibré).
        """
        if cible:
            try:
                member = tools.member(cible)
            except ValueError:
                await ctx.send("Cible introuvable.")
                return
        else:
            member = ctx.author

        await features.inscription.main(member)


    @commands.command()
    @tools.mjs_only
    async def doas(self, ctx, *, qui_quoi):
        """Exécute une commande en tant qu'un autre joueur (COMMANDE MJ)

        Args:
            qui_quoi: nom de la cible (nom/mention d'un joueur INSCRIT)
                suivi de la commande à exécuter (commençant par un ``!``).

        Example:
            ``!doas Vincent Croquette !vote Annie Colin``
        """
        sep = " " + config.bot.command_prefix
        qui, _, quoi = qui_quoi.partition(sep)
        # !doas <@!id> !vote R ==> qui = "<@!id>", quoi = "vote R"
        if not quoi:
            raise commands.UserInputError(f"'{sep}' not found in qui_quoi")
        joueur = await tools.boucle_query_joueur(ctx, qui.strip())

        ctx.message.content = config.bot.command_prefix + quoi
        try:
            member = joueur.member
        except ValueError:
            await ctx.send(f"{joueur} absent du serveur, "
                           "tentative de contournement")
            class PseudoMember:
                __class__ = discord.Member
                id = joueur.discord_id
                display_name = joueur.nom
                guild = config.guild
                mention = f"[@{joueur.nom}]"
                top_role = (config.Role.joueur_en_vie if joueur.est_vivant
                            else config.Role.joueur_mort)
                roles = [config.Role.everyone, top_role]
            member = PseudoMember()

        ctx.message.author = member

        await ctx.send(f":robot: Exécution en tant que {joueur.nom} :")
        with one_command.bypass(ctx):
            await config.bot.process_commands(ctx.message)


    @commands.command(aliases=["autodestruct", "ad"])
    @tools.mjs_only
    async def secret(self, ctx, *, quoi):
        """Supprime le message puis exécute la commande (COMMANDE MJ)

        Args:
            quoi: commande à exécuter, commençant par un ``!``

        Utile notemment pour faire des commandes dans un channel public,
        pour que la commande (moche) soit immédiatement supprimée.
        """
        await ctx.message.delete()

        ctx.message.content = quoi

        with one_command.bypass(ctx):
            await config.bot.process_commands(ctx.message)


    @one_command.do_not_limit
    @commands.command()
    @tools.private
    async def stop(self, ctx):
        """Peut débloquer des situations compliquées (beta)

        Ne pas utiliser cette commande sauf en cas de force majeure
        où plus rien ne marche, et sur demande d'un MJ (après c'est
        pas dit que ça marche mieux après l'avoir utilisée)
        """
        if ctx.channel.id in config.bot.in_command:
            config.bot.in_command.remove(ctx.channel.id)
        await ctx.send("Te voilà libre, camarade !")


    @commands.command(aliases=["aide", "aled", "oskour"])
    async def help(self, ctx, *, command=None):
        """Affiche la liste des commandes utilisables et leur utilisation

        Args:
            command (optionnel): nom exact d'une commande à expliquer
                (ou un de ses alias)

        Si ``command`` n'est pas précisée, liste l'ensemble des commandes
        accessibles à l'utilisateur.
        """
        pref = config.bot.command_prefix
        cogs = config.bot.cogs                  # Dictionnaire nom: cog
        commandes = {cmd.name: cmd for cmd in config.bot.commands}
        aliases = {alias: nom for nom, cmd in commandes.items()
                   for alias in cmd.aliases}
        # Dictionnaire alias: nom de la commande

        len_max = max(len(cmd) for cmd in commandes)

        def descr_command(cmd):
            return f"\n  - {pref}{cmd.name.ljust(len_max)}  {cmd.short_doc}"

        if not command:
            # Pas d'argument ==> liste toutes les commandes

            r = f"{config.bot.description} (v{__version__})"
            for cog in cogs.values():
                runnables = await _filter_runnables(cog.get_commands(), ctx)
                if not runnables:
                    # pas de runnables dans le cog, on passe
                    continue

                r += f"\n\n{type(cog).__name__} - {cog.description} :"
                for cmd in runnables:       # pour chaque commande runnable
                    r += descr_command(cmd)

            runnables_hors_cog = await _filter_runnables(
                (cmd for cmd in config.bot.commands if not cmd.cog), ctx
            )
            if runnables_hors_cog:
                r += "\n\nCommandes isolées :"
                for cmd in runnables_hors_cog:
                    r += descr_command(cmd)

            r += (f"\n\nUtilise <{pref}help command> pour "
                  "plus d'information sur une commande.")

        else:
            # Aide détaillée sur une commande
            if command.startswith(pref):
                # Si le joueur fait !help !command (ou !help ! command)
                command = command.lstrip(pref).strip()
            if command in aliases:
                # Si !help d'un alias
                command = aliases[command]

            if command in commandes:            # Si commande existante
                cmd = commandes[command]

                doc = cmd.help or ""
                doc = doc.replace("``", "`")
                doc = doc.replace("Args:", "Arguments :")
                doc = doc.replace("Warning:", "Avertissement :")
                doc = doc.replace("Examples:", "Exemples :")
                doc = re.sub(r":\w+?:`[\.~!]*(.+?)`", r"`\1`", doc)
                # enlève les :class: et consors

                if isinstance(cmd, commands.Group):
                    r = (f"{pref}{command} <option> [args...] – {doc}\n\n"
                         "Options :\n")

                    scommands = sorted(cmd.commands, key=lambda cmd: cmd.name)
                    options = [f"{scmd.name} {scmd.signature}"
                               for scmd in scommands]
                    slen_max = max(len(opt) for opt in options)
                    r += "\n".join(f"    - {pref}{command} "
                                   f"{opt.ljust(slen_max)}  {scmd.short_doc}"
                                   for scmd, opt in zip(scommands, options))
                else:
                    r = f"{pref}{command} {cmd.signature} – {doc}"

                if cmd.aliases:         # Si la commande a des alias
                    r += f"\n\nAlias : {pref}" + f", {pref}".join(cmd.aliases)

            else:
                r = (f"Commande '{pref}{command}' non trouvée.\n"
                     f"Utilise '{pref}help' pour la liste des commandes.")

        r += ("\n\nSi besoin, n'hésite pas à appeler un MJ "
              "en les mentionnant (@MJ).")

        await tools.send_code_blocs(ctx, r, sep="\n\n")
        # On envoie, en séparant enntre les cogs de préférence


    @commands.command(aliases=["about", "copyright", "licence", "auteurs"])
    async def apropos(self, ctx):
        """Informations et mentions légales du projet

        N'hésitez-pas à nous contacter pour en savoir plus !
        """
        embed = discord.Embed(
            title=f"**LG-bot** - v{__version__}",
            description=config.bot.description
        ).set_author(
            name="À propos de ce bot :",
            icon_url=config.bot.user.avatar_url,
        ).set_image(
            url=("https://gist.githubusercontent.com/loic-simon/"
                 "66c726053323017dba67f85d942495ef/raw/"
                 "48f2607a61f3fc1b7285fd64873621035c6fbbdb/logo_espci.png"),
        ).add_field(
            name="Auteurs",
            value="Loïc Simon\nTom Lacoma",
            inline=True,
        ).add_field(
            name="Licence",
            value="Projet open-source sous licence MIT\n"
                  "https://opensource.org/licenses/MIT",
            inline=True,
        ).add_field(
            name="Pour en savoir plus :",
            value="https://github.com/loic-simon/lg-rez",
            inline=False,
        ).add_field(
            name="Copyright :",
            value=":copyright: 2021 Club BD-Jeux × GRIs – ESPCI Paris - PSL",
            inline=False,
        ).set_footer(
            text="Retrouvez-nous sur Discord : LaCarpe#1674, TaupeOrAfk#3218",
        )
        await ctx.send(embed=embed)



import datetime
import unittest
from unittest import mock

import discord
import freezegun

from lgrez import config, bdd
from lgrez.features import inscription
from lgrez.blocs import gsheets
from test import mock_discord, mock_bdd, mock_env


class TestInscriptionFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.inscription functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()


    @mock.patch("lgrez.blocs.tools.channel")
    async def test_new_channel(self, ch_patch):
        """Unit tests for inscription.new_channel function."""
        # async def new_channel(member)
        new_channel = inscription.new_channel
        categ = mock.MagicMock(discord.CategoryChannel)

        # < 50 channels
        categ.channels = [0]*49
        ch_patch.return_value = categ
        member = mock.MagicMock(discord.Member)
        member.guild.create_text_channel = mock.AsyncMock()
        chan = await new_channel(member)
        ch_patch.assert_called_once()
        member.guild.create_text_channel.assert_called_once_with(
            f"{config.private_chan_prefix}{member.name}",
            topic=str(member.id), category=categ
        )
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        ch_patch.reset_mock()

        # 50 <= channels < 100, cat 2 déjà existante
        categ.channels = [0]*50
        categ2 = mock.MagicMock(discord.CategoryChannel)
        categ2.channels = [0]*49
        ch_patch.side_effect = [categ, categ2]
        member = mock.MagicMock(discord.Member)
        member.guild.create_text_channel = mock.AsyncMock()
        chan = await new_channel(member)
        ch_patch.assert_called()
        self.assertEqual(ch_patch.call_count, 2)
        categ.clone.assert_not_called()
        self.assertEqual([call.args[0] for call in ch_patch.call_args_list],
                         [f"{config.private_chan_category_name}",
                          f"{config.private_chan_category_name} 2"])
        member.guild.create_text_channel.assert_called_once_with(
            f"{config.private_chan_prefix}{member.name}",
            topic=str(member.id), category=categ2
        )
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        ch_patch.reset_mock()

        # 50 <= channels < 100, cat 2 non existante
        ch_patch.side_effect = [categ, None]
        member = mock.MagicMock(discord.Member)
        member.guild.create_text_channel = mock.AsyncMock()
        chan = await new_channel(member)
        ch_patch.assert_called()
        self.assertEqual(ch_patch.call_count, 2)
        categ.clone.assert_called_once_with(
            name=f"{config.private_chan_category_name} 2"
        )
        self.assertEqual([call.args[0] for call in ch_patch.call_args_list],
                         [f"{config.private_chan_category_name}",
                          f"{config.private_chan_category_name} 2"])
        member.guild.create_text_channel.assert_called_once_with(
            f"{config.private_chan_prefix}{member.name}",
            topic=str(member.id), category=categ.clone.return_value
        )
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        ch_patch.reset_mock()
        categ.clone.reset_mock()

        # 100 <= channels, cat 3 déjà existante
        categ2.channels = [0]*50
        categ3 = mock.MagicMock(discord.CategoryChannel)
        categ3.channels = [0]*49
        ch_patch.side_effect = [categ, categ2, categ3]
        member = mock.MagicMock(discord.Member)
        member.guild.create_text_channel = mock.AsyncMock()
        chan = await new_channel(member)
        ch_patch.assert_called()
        self.assertEqual(ch_patch.call_count, 3)
        categ.clone.assert_not_called()
        self.assertEqual([call.args[0] for call in ch_patch.call_args_list],
                         [f"{config.private_chan_category_name}",
                          f"{config.private_chan_category_name} 2",
                          f"{config.private_chan_category_name} 3"])
        member.guild.create_text_channel.assert_called_once_with(
            f"{config.private_chan_prefix}{member.name}",
            topic=str(member.id), category=categ3
        )
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        ch_patch.reset_mock()

        # 100 <= channels, cat 3 non existante
        ch_patch.side_effect = [categ, categ2, None]
        member = mock.MagicMock(discord.Member)
        member.guild.create_text_channel = mock.AsyncMock()
        chan = await new_channel(member)
        ch_patch.assert_called()
        self.assertEqual(ch_patch.call_count, 3)
        categ.clone.assert_called_once_with(
            name=f"{config.private_chan_category_name} 3"
        )
        self.assertEqual([call.args[0] for call in ch_patch.call_args_list],
                         [f"{config.private_chan_category_name}",
                          f"{config.private_chan_category_name} 2",
                          f"{config.private_chan_category_name} 3"])
        member.guild.create_text_channel.assert_called_once_with(
            f"{config.private_chan_prefix}{member.name}",
            topic=str(member.id), category=categ.clone.return_value
        )
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )


    @mock_env.patch_env(LGREZ_TDB_SHEET_ID="bzoulip!")
    @mock.patch("lgrez.blocs.gsheets.connect")
    @mock.patch("lgrez.blocs.gsheets.update")
    async def test_register_on_tdb(self, gupdate_patch, gconnect_patch):
        """Unit tests for inscription.register_on_tdb function."""
        # def register_on_tdb(joueur)
        register_on_tdb = inscription.register_on_tdb
        config.tdb_main_sheet = "uéué"
        config.tdb_header_row = 7
        config.tdb_id_column = "C"
        config.tdb_main_columns = ("O", "V")
        config.tdb_tampon_columns = ("E", "L")

        jr = mock.Mock()

        # mauvaise colonne primaire
        values = [0]*6 + [[0, 0, "bizzp"]]
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        with self.assertRaises(ValueError) as cm:
            register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        self.assertIn("`C7` vaut `bizzp`", cm.exception.args[0])
        self.assertIn(bdd.Joueur.primary_col.key, cm.exception.args[0])
        gupdate_patch.assert_not_called()

        # mauvaise colonne
        values = [0]*6 + [[0, 0, bdd.Joueur.primary_col.key, "d", "e", "f",
                           "g", "h", "i", "j", "k", "l", "m", "n", "koko"]]
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        with self.assertRaises(ValueError) as cm:
            register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        self.assertIn("`koko` n'est pas une colonne", cm.exception.args[0])
        gupdate_patch.assert_not_called()

        # mauvais tampon
        values = [0]*6 + [[0, 0, bdd.Joueur.primary_col.key, "d", "koko",
                           "f", "g", "h", "i", "j", "k", "l", "m", "n", "nom",
                           "chambre", "statut", "role", "camp",
                           "votant_village", "votant_loups", "role_actif",
                            "koko"]]
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        with self.assertRaises(ValueError) as cm:
            register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        self.assertIn("koko", cm.exception.args[0])
        self.assertIn("n'est pas une colonne", cm.exception.args[0])
        gupdate_patch.assert_not_called()

        # aucun joueur inscrit sur le TDB
        values = [0]*6 + [
            [0, 0, bdd.Joueur.primary_col.key, "d", "tampon_nom",
                "tampon_chambre", "tampon_statut", "tampon_role",
                "tampon_camp", "tampon_votant_village", "tampon_votant_loups",
                "tampon_role_actif", "m", "n", "nom", "chambre", "statut",
                "role", "camp", "votant_village", "votant_loups", "role_actif",
                "koko"],
        ]
        base = values
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        gupdate_patch.assert_called_once()
        usheet, *modifs = gupdate_patch.call_args.args
        self.assertEqual(usheet, sheet)
        mods = {
            gsheets.Modif(7, 2, jr.discord_id),
            gsheets.Modif(7, 4, jr.nom), gsheets.Modif(7, 14, jr.nom),
            gsheets.Modif(7, 5, jr.chambre), gsheets.Modif(7, 15, jr.chambre),
            gsheets.Modif(7, 6, jr.statut), gsheets.Modif(7, 16, jr.statut),
            gsheets.Modif(7, 7, jr.role), gsheets.Modif(7, 17, jr.role),
            gsheets.Modif(7, 8, jr.camp), gsheets.Modif(7, 18, jr.camp),
            gsheets.Modif(7, 9, jr.votant_village),
            gsheets.Modif(7, 19, jr.votant_village),
            gsheets.Modif(7, 10, jr.votant_loups),
            gsheets.Modif(7, 20, jr.votant_loups),
            gsheets.Modif(7, 11, jr.role_actif),
            gsheets.Modif(7, 21, jr.role_actif),
        }
        self.assertEqual(mods, set(modifs))
        gupdate_patch.reset_mock()

        # junk en dessous du header
        values = base + [
            [0, 0, "brzzz"],
            [0, 0, "brozozoz"],
        ]
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        gupdate_patch.assert_called_once()
        usheet, *modifs = gupdate_patch.call_args.args
        self.assertEqual(usheet, sheet)
        mods = {
            gsheets.Modif(7, 2, jr.discord_id),
            gsheets.Modif(7, 4, jr.nom), gsheets.Modif(7, 14, jr.nom),
            gsheets.Modif(7, 5, jr.chambre), gsheets.Modif(7, 15, jr.chambre),
            gsheets.Modif(7, 6, jr.statut), gsheets.Modif(7, 16, jr.statut),
            gsheets.Modif(7, 7, jr.role), gsheets.Modif(7, 17, jr.role),
            gsheets.Modif(7, 8, jr.camp), gsheets.Modif(7, 18, jr.camp),
            gsheets.Modif(7, 9, jr.votant_village),
            gsheets.Modif(7, 19, jr.votant_village),
            gsheets.Modif(7, 10, jr.votant_loups),
            gsheets.Modif(7, 20, jr.votant_loups),
            gsheets.Modif(7, 11, jr.role_actif),
            gsheets.Modif(7, 21, jr.role_actif),
        }
        self.assertEqual(mods, set(modifs))
        gupdate_patch.reset_mock()

        # 3 joueurs inscrits
        values = base + [
            [0, 0, "1351"],
            [0, 0, "313"],
            [0, 0, "1101"],
            [0, 0, "junk"],
        ]
        sheet = mock.Mock(**{"get_all_values.return_value": values})
        workbook = mock.Mock(**{"worksheet.return_value": sheet})
        gconnect_patch.return_value = workbook
        register_on_tdb(jr)
        gconnect_patch.assert_called_once_with("bzoulip!")
        workbook.worksheet.assert_called_once_with("uéué")
        gconnect_patch.reset_mock()
        gupdate_patch.assert_called_once()
        usheet, *modifs = gupdate_patch.call_args.args
        self.assertEqual(usheet, sheet)
        mods = {
            gsheets.Modif(10, 2, jr.discord_id),
            gsheets.Modif(10, 4, jr.nom), gsheets.Modif(10, 14, jr.nom),
            gsheets.Modif(10, 5, jr.chambre),
            gsheets.Modif(10, 15, jr.chambre),
            gsheets.Modif(10, 6, jr.statut), gsheets.Modif(10, 16, jr.statut),
            gsheets.Modif(10, 7, jr.role), gsheets.Modif(10, 17, jr.role),
            gsheets.Modif(10, 8, jr.camp), gsheets.Modif(10, 18, jr.camp),
            gsheets.Modif(10, 9, jr.votant_village),
            gsheets.Modif(10, 19, jr.votant_village),
            gsheets.Modif(10, 10, jr.votant_loups),
            gsheets.Modif(10, 20, jr.votant_loups),
            gsheets.Modif(10, 11, jr.role_actif),
            gsheets.Modif(10, 21, jr.role_actif),
        }
        self.assertEqual(mods, set(modifs))


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.inscription.new_channel")
    @mock.patch("lgrez.features.inscription.register_on_tdb")
    @mock.patch("lgrez.blocs.tools.sleep")
    async def test_main(self, sleep_patch, reg_patch, newchan_patch):
        """Unit tests for inscription.main function."""
        # async def main(member)
        main = inscription.main
        mock_bdd.add_campsroles()

        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur 1")
        joueur1.add()

        newchan_patch.return_value = mock_discord.chan("bzzt", id=798)

        # Joueur inscrit en base
        with mock_discord.mock_members_and_chans(joueur1):
            member = joueur1.member
            chan = joueur1.private_chan
            await main(member)
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        mock_discord.assert_sent(chan, f"{member.mention} tu es déjà inscrit")
        reg_patch.assert_not_called()
        newchan_patch.assert_not_called()
        sleep_patch.assert_not_called()
        joueur1.delete()

        # Joueur en cours d'inscription -> abort
        chan = mock_discord.chan("bzz", topic="123456")
        config.guild.text_channels = [chan]
        member = mock.Mock(discord.Member, id=123456)
        with mock_discord.interact(("yes_no", False)):
            await main(member)
        chan.set_permissions.assert_called_once_with(
            member, read_messages=True, send_messages=True
        )
        mock_discord.assert_sent(chan, f"{member.mention}, par ici",
                                 f"Bienvenue {member.mention}",
                                 f"finalisons ton inscription",
                                 f"C'est bon pour toi ?",
                                 "Pas de soucis")
        reg_patch.assert_not_called()
        newchan_patch.assert_not_called()
        sleep_patch.assert_called()
        sleep_patch.reset_mock()

        # Joueur arrivant -> abort
        member = mock.Mock(discord.Member, id=145)
        with mock_discord.interact(("yes_no", False)):
            await main(member)
        newchan_patch.assert_called_once_with(member)
        chan = newchan_patch.return_value
        mock_discord.assert_sent(chan, f"Bienvenue {member.mention}",
                                 f"finalisons ton inscription",
                                 f"C'est bon pour toi ?",
                                 "Pas de soucis")
        reg_patch.assert_not_called()
        newchan_patch.reset_mock()
        sleep_patch.assert_called()
        sleep_patch.reset_mock()

        # Joueur arrivant -> ok -> name -> troll -> good name
        # demande_chambre False
        config.demande_chambre = False
        member = mock.Mock(discord.Member, id=145)
        member.top_role.__lt__ = lambda s, o: True  # role < MJ
        member.display_name = "Pr-Z N0 N"
        with mock_discord.interact(
            ("yes_no", True),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="préz")),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="no n")),
            ("yes_no", False),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="pr-z")),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="n0 n")),
            ("yes_no", True),
        ):
            await main(member)
        newchan_patch.assert_called_once_with(member)
        chan = newchan_patch.return_value
        mock_discord.assert_sent(
            chan, f"Bienvenue {member.mention}", "finalisons ton inscription",
            f"C'est bon pour toi ?", ["Parfait", "prénom"], "prénom",
            "nom de famille", "Préz No N", "prénom", "nom de famille",
            "Pr-Z N0 N", "Je t'ai renommé", "Je t'inscris",
            "Tu es maintenant inscrit", "quelques dernières choses",
            "c'est tout bon"
        )
        chan.edit.assert_any_call(
            name=f"{config.private_chan_prefix}Pr-Z N0 N"
        )
        member.edit.assert_called_once_with(nick="Pr-Z N0 N")
        member.edit.reset_mock()
        self.assertEqual(len(bdd.Joueur.query.all()), 1)
        jr = bdd.Joueur.query.one()
        self.assertEqual(
            [jr.discord_id, jr.chan_id_, jr.nom, jr.chambre, jr.statut,
             jr.role, jr.camp, jr.votant_village, jr.votant_loups,
             jr.role_actif],
            [145, 798, "Pr-Z N0 N", None, bdd.Statut.vivant,
             bdd.Role.default(), bdd.Camp.default(), True, False, False]
        )
        member.add_roles.assert_called_once_with(config.Role.joueur_en_vie)
        chan.edit.assert_any_call(topic=mock.ANY)
        reg_patch.assert_called_once_with(jr)
        reg_patch.reset_mock()
        newchan_patch.reset_mock()
        sleep_patch.assert_called()
        sleep_patch.reset_mock()
        jr.delete()

        # Joueur arrivant -> ok -> name -> chambre -> pas à la Rez
        config.demande_chambre = True
        member = mock.Mock(discord.Member, id=145)
        member.top_role.__lt__ = lambda s, o: True  # role < MJ
        member.display_name = "Pr-Z N0 N"
        with mock_discord.interact(
            ("yes_no", True),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="pr-z")),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="n0 n")),
            ("yes_no", True),
            ("yes_no", False),
        ):
            await main(member)
        newchan_patch.assert_called_once_with(member)
        chan = newchan_patch.return_value
        mock_discord.assert_sent(
            chan, f"Bienvenue {member.mention}", "finalisons ton inscription",
            f"C'est bon pour toi ?", ["Parfait", "prénom"], "prénom",
            "nom de famille",
            "Pr-Z N0 N", "Je t'ai renommé", "habites-tu à la Rez",
            "Je t'inscris", "Tu es maintenant inscrit",
            "quelques dernières choses", "c'est tout bon"
        )
        chan.edit.assert_any_call(
            name=f"{config.private_chan_prefix}Pr-Z N0 N"
        )
        member.edit.assert_called_once_with(nick="Pr-Z N0 N")
        member.edit.reset_mock()
        self.assertEqual(len(bdd.Joueur.query.all()), 1)
        jr = bdd.Joueur.query.one()
        self.assertEqual(
            [jr.discord_id, jr.chan_id_, jr.nom, jr.chambre, jr.statut,
             jr.role, jr.camp, jr.votant_village, jr.votant_loups,
             jr.role_actif],
            [145, 798, "Pr-Z N0 N", config.chambre_mj, bdd.Statut.vivant,
             bdd.Role.default(), bdd.Camp.default(), True, False, False]
        )
        member.add_roles.assert_called_once_with(config.Role.joueur_en_vie)
        chan.edit.assert_any_call(topic=mock.ANY)
        reg_patch.assert_called_once_with(jr)
        reg_patch.reset_mock()
        newchan_patch.reset_mock()
        sleep_patch.assert_called()
        sleep_patch.reset_mock()
        jr.delete()

        # Joueur arrivant -> ok -> name -> chambre -> à la Rez
        config.demande_chambre = True
        member = mock.Mock(discord.Member, id=145)
        member.top_role.__lt__ = lambda s, o: True  # role < MJ
        member.display_name = "Pr-Z N0 N"
        with mock_discord.interact(
            ("yes_no", True),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="pr-z")),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="n0 n")),
            ("yes_no", True),
            ("yes_no", True),
            ("wait_for_message", mock.NonCallableMock(discord.Message,
                                                      content="214")),
        ):
            await main(member)
        newchan_patch.assert_called_once_with(member)
        chan = newchan_patch.return_value
        mock_discord.assert_sent(
            chan, f"Bienvenue {member.mention}", "finalisons ton inscription",
            f"C'est bon pour toi ?", ["Parfait", "prénom"], "prénom",
            "nom de famille",
            "Pr-Z N0 N", "Je t'ai renommé", "habites-tu à la Rez",
            "quelle est ta chambre", "Je t'inscris",
            "Tu es maintenant inscrit", "quelques dernières choses",
            "c'est tout bon"
        )
        chan.edit.assert_any_call(
            name=f"{config.private_chan_prefix}Pr-Z N0 N"
        )
        member.edit.assert_called_once_with(nick="Pr-Z N0 N")
        member.edit.reset_mock()
        self.assertEqual(len(bdd.Joueur.query.all()), 1)
        jr = bdd.Joueur.query.one()
        self.assertEqual(
            [jr.discord_id, jr.chan_id_, jr.nom, jr.chambre, jr.statut,
             jr.role, jr.camp, jr.votant_village, jr.votant_loups,
             jr.role_actif],
            [145, 798, "Pr-Z N0 N", "214", bdd.Statut.vivant,
             bdd.Role.default(), bdd.Camp.default(), True, False, False]
        )
        member.add_roles.assert_called_once_with(config.Role.joueur_en_vie)
        chan.edit.assert_any_call(topic=mock.ANY)
        reg_patch.assert_called_once_with(jr)
        reg_patch.reset_mock()
        newchan_patch.reset_mock()
        sleep_patch.assert_called()
        sleep_patch.reset_mock()
        jr.delete()



import unittest
from unittest import mock

from lgrez import config, bdd
from lgrez.features import actions_publiques
from test import mock_discord, mock_bdd, mock_env


class TestActionsPubliques(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.actions_publiques commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = actions_publiques.ActionsPubliques(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.config.Channel.haros.send")
    @mock.patch("lgrez.config.Channel.debats.send")
    @mock_discord.interact()
    async def test_haro(self, send_debats_patch, send_haros_patch):
        """Unit tests for !haro command."""
        # async def haro(self, ctx, *, cible=None)
        haro = self.cog.haro
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            vote_condamne_=None)
        joueur.add()
        cible = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2")
        cible.add()

        # no vote
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("Pas de vote")
        self.assertFalse(bdd.CandidHaro.query.all())    # no haro created
        send_debats_patch.assert_not_called()
        send_haros_patch.assert_not_called()

        # cible = dead
        joueur.vote_condamne_ = "bzzzt"
        cible.statut = bdd.Statut.mort
        cible.update()
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("pas assez souffert")
        self.assertFalse(bdd.CandidHaro.query.all())    # no haro created
        send_debats_patch.assert_not_called()
        send_haros_patch.assert_not_called()

        # cible = immortel
        cible.statut = bdd.Statut.immortel
        cible.update()
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("Comment oses-tu")
        self.assertFalse(bdd.CandidHaro.query.all())    # no haro created
        send_debats_patch.assert_not_called()
        send_haros_patch.assert_not_called()

        # cible = MV : ok
        cible.statut = bdd.Statut.MV
        cible.update()

        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message(".")),
                ("yes_no", False)):
            await ctx.invoke()

        ctx.assert_sent("quelle est la raison", "", "")
        self.assertFalse(bdd.CandidHaro.query.all())    # no haro created
        send_debats_patch.assert_not_called()
        send_haros_patch.assert_not_called()

        # ok, and abort
        cible.statut = bdd.Statut.vivant
        cible.update()

        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("pakontan")),
                ("yes_no", False)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("quelle est la raison",
                        "tout bon ?",
                        "Mission aborted")
        embed = ctx.send.call_args_list[1].kwargs["embed"]
        self.assertIn("contre Joueur2", embed.title)
        self.assertIn("pakontan", embed.description)
        self.assertIn("Joueur1", embed.author.name)
        self.assertFalse(bdd.CandidHaro.query.all())    # no haro created
        send_debats_patch.assert_not_called()
        send_haros_patch.assert_not_called()

        # validate
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("pakontan")),
                ("yes_no", True)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("quelle est la raison",
                        "tout bon ?",
                        "c'est parti")
        embed = ctx.send.call_args_list[1].kwargs["embed"]
        haros = bdd.CandidHaro.query.all()
        self.assertEqual(len(haros), 2)    # 2 haros created
        self.assertEqual({haro.joueur for haro in haros}, {joueur, cible})
        self.assertEqual(haros[0].type, bdd.CandidHaroType.haro)
        self.assertEqual(haros[1].type, bdd.CandidHaroType.haro)
        send_haros_patch.assert_called_once()
        self.assertEqual(send_haros_patch.call_args.kwargs, {"embed": embed})
        send_haros_patch.reset_mock()
        send_debats_patch.assert_called_once()
        send_debats_patch.reset_mock()

        # validate, but haroted already registered
        bdd.CandidHaro.delete(*haros)
        self.assertFalse(bdd.CandidHaro.query.all())    # no haros
        bdd.CandidHaro(joueur=cible, type=bdd.CandidHaroType.haro).add()
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("pakontan")),
                ("yes_no", True)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("", "", "")
        haros = bdd.CandidHaro.query.all()
        self.assertEqual(len(haros), 2)    # 2 haros created
        self.assertEqual({haro.joueur for haro in haros}, {joueur, cible})
        self.assertEqual(haros[0].type, bdd.CandidHaroType.haro)
        self.assertEqual(haros[1].type, bdd.CandidHaroType.haro)
        send_haros_patch.assert_called_once()
        send_haros_patch.reset_mock()
        send_debats_patch.assert_called_once()
        send_debats_patch.reset_mock()

        # validate, but haroter already registered
        bdd.CandidHaro.delete(*haros)
        self.assertFalse(bdd.CandidHaro.query.all())    # no haros
        bdd.CandidHaro(joueur=joueur, type=bdd.CandidHaroType.haro).add()
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("pakontan")),
                ("yes_no", True)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("", "", "")
        haros = bdd.CandidHaro.query.all()
        self.assertEqual(len(haros), 2)    # 2 haros created
        self.assertEqual({haro.joueur for haro in haros}, {joueur, cible})
        self.assertEqual(haros[0].type, bdd.CandidHaroType.haro)
        self.assertEqual(haros[1].type, bdd.CandidHaroType.haro)
        send_haros_patch.assert_called_once()
        send_haros_patch.reset_mock()
        send_debats_patch.assert_called_once()
        send_debats_patch.reset_mock()

        # validate, but haroter and haroted already registered
        bdd.CandidHaro.delete(*haros)
        self.assertFalse(bdd.CandidHaro.query.all())    # no haros
        bdd.CandidHaro(joueur=joueur, type=bdd.CandidHaroType.haro).add()
        bdd.CandidHaro(joueur=cible, type=bdd.CandidHaroType.haro).add()
        ctx = mock_discord.get_ctx(haro, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("pakontan")),
                ("yes_no", True)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("", "", "")
        haros = bdd.CandidHaro.query.all()
        self.assertEqual(len(haros), 2)    # 2 haros created
        self.assertEqual({haro.joueur for haro in haros}, {joueur, cible})
        self.assertEqual(haros[0].type, bdd.CandidHaroType.haro)
        self.assertEqual(haros[1].type, bdd.CandidHaroType.haro)
        send_haros_patch.assert_called_once()
        send_haros_patch.reset_mock()
        send_debats_patch.assert_called_once()
        send_debats_patch.reset_mock()



    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.config.Channel.haros.send")
    @mock.patch("lgrez.config.Channel.debats.send")
    @mock_discord.interact()
    async def test_candid(self, send_debats_patch, send_candids_patch):
        """Unit tests for !candid command."""
        # async def candid(self, ctx)
        candid = self.cog.candid
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            vote_maire_=None)
        joueur.add()

        # no vote
        ctx = mock_discord.get_ctx(candid, _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("Pas de vote")
        self.assertFalse(bdd.CandidHaro.query.all())    # no candid created
        send_debats_patch.assert_not_called()
        send_candids_patch.assert_not_called()

        # joueur already candidated
        joueur.vote_maire_ = "bzzzt"
        cand = bdd.CandidHaro(joueur=joueur,
                                type=bdd.CandidHaroType.candidature)
        cand.add()
        ctx = mock_discord.get_ctx(candid, _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("déjà présenté")
        self.assertEqual(len(bdd.CandidHaro.query.all()), 1)    # none created
        send_debats_patch.assert_not_called()
        send_candids_patch.assert_not_called()

        # ok, and abort
        cand.delete()
        ctx = mock_discord.get_ctx(candid, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("votépourmoi")),
                ("yes_no", False)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("Quel est ton programme",
                        "tout bon ?",
                        "Mission aborted")
        embed = ctx.send.call_args_list[1].kwargs["embed"]
        self.assertIn("Joueur1", embed.title)
        self.assertIn("votépourmoi", embed.description)
        self.assertIn("Joueur1", embed.author.name)
        self.assertFalse(bdd.CandidHaro.query.all())    # no candid created
        send_debats_patch.assert_not_called()
        send_candids_patch.assert_not_called()

        # validate
        ctx = mock_discord.get_ctx(candid, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("votépourmoi")),
                ("yes_no", True)):
            # give reason then abort (answer "no" at check)
            await ctx.invoke()
        ctx.assert_sent("Quel est ton programme",
                        "tout bon ?",
                        "c'est parti")
        embed = ctx.send.call_args_list[1].kwargs["embed"]
        candids = bdd.CandidHaro.query.all()
        self.assertEqual(len(candids), 1)    # 2 candids created
        self.assertEqual(candids[0].joueur, joueur)
        self.assertEqual(candids[0].type, bdd.CandidHaroType.candidature)
        send_candids_patch.assert_called_once()
        self.assertEqual(send_candids_patch.call_args.kwargs, {"embed": embed})
        send_candids_patch.reset_mock()
        send_debats_patch.assert_called_once()
        send_debats_patch.reset_mock()



"""lg-rez / features / Gestion des votes et actions

Ouverture / fermeture / rappels des votes et actions (+ refill)

"""

import datetime

from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.features import gestion_actions
from lgrez.bdd import (Joueur, Action, BaseAction, Tache, CandidHaro,
                       Utilisation, CandidHaroType, ActionTrigger,
                       UtilEtat, Vote)


async def recup_joueurs(quoi, qui, heure=None):
    """Renvoie les joueurs concernés par la tâche !quoi <qui> [heure].

    Args:
        quoi (str): évènement, ``"open" / "close" / "remind"``.
        qui (:class:`.bdd.Vote` | :class:`str`):
            ===========     ===========
            ``Vote``        pour le vote correspondant
            ``action``      pour les actions commençant à ``heure``
            ``{id}``        pour une action précise (:attr:`bdd.Action.id`)
            ===========     ===========

        heure (str): si ``qui == "action"``, heure associée
            (au format ``HHhMM``).

    Returns:
        :class:`list`\[:class:`.bdd.Joueur`\]

    Examples:
        ``!open cond`` -> joueurs avec droit de vote
        ``!close action 17h`` -> joueurs dont l'action se termine à 17h
    """
    if quoi not in ["open", "close", "remind"]:
        raise ValueError(f"recup_joueurs: bad value for `quoi`: `{quoi}`")

    if isinstance(qui, Vote):
        # Critère principale : présence/absence d'une action actuellement
        # ouverte (et non traitée pour remind)
        criteres = {
            "open": ~Joueur.actions.any(Action.is_open, vote=qui),
            "close": Joueur.actions.any(Action.is_open, vote=qui),
            "remind": Joueur.actions.any(Action.is_waiting, vote=qui),
        }
        critere = criteres[quoi]
        if quoi == "open":
            # Open : le joueur doit en plus avoir votant_village/loups True
            if qui == Vote.loups:
                critere &= Joueur.votant_loups.is_(True)
            else:
                critere &= Joueur.votant_village.is_(True)

        return Joueur.query.filter(critere).all()
        # Liste des joueurs répondant aux critères

    elif qui == "action":
        if heure and isinstance(heure, str):
            # Si l'heure est précisée, on convertit "HHhMM" -> datetime.time
            tps = tools.heure_to_time(heure)
        else:
            raise commands.BadArgument(
                "[heure] doit être spécifiée lorque <qui> == \"action\""
            )

        actions = gestion_actions.get_actions(
            quoi, ActionTrigger.temporel, tps
        )

        dic = {}
        for action in actions:
            joueur = action.joueur
            if joueur not in dic:
                dic[joueur] = []
            dic[joueur].append(action)

        return dic
        # Formerly :
        # {joueur.player_id:[action for action in actions if
        # action.player_id == joueur.player_id] for joueur in
        # [Joueur.query.get(action.player_id) for action in actions]}

    elif qui.isdigit():
        action = Action.query.get(int(qui))
        if not action:
            raise commands.BadArgument(f"Pas d'action d'ID = {qui}")
        if not action.active:
            raise commands.BadArgument(f"Action d'ID = {qui} inactive")

        # Appel direct action par son numéro (perma : rappel seulement)
        if ((quoi == "open" and (
                not action.is_open
                or action.base.trigger_debut == ActionTrigger.perma
            ))
            or (quoi == "close" and action.is_open)
            or (quoi == "remind" and action.is_waiting)):
            # Action lançable
            return {action.joueur: [action]}
        else:
            return {}

    else:
        raise commands.BadArgument(f"""Argument <qui> == \"{qui}" invalide""")


async def _do_refill(motif, actions):
    # Détermination nouveau nombre de charges
    if motif in config.refills_full:
        # Refill -> nombre de charges initial de l'action
        new_charges = {action: action.base.base_charges for action in actions}
    else:
        # Refill -> + 1 charge
        new_charges = {action: action.charges + 1 for action in actions}

    # Refill proprement dit
    for action, charge in new_charges.items():
        if charge <= action.charges:
            # Pas de rechargement à faire (déjà base_charges)
            continue

        if (not action.charges
            and action.base.trigger_debut == ActionTrigger.perma):
            # Action permanente qui était épuisée : on ré-ouvre !
            if tools.en_pause():
                ts = tools.fin_pause()
            else:
                ts = datetime.datetime.now() + datetime.timedelta(seconds=10)
                # + 10 secondes pour ouvrir après le message de refill
            Tache(timestamp=ts,
                  commande=f"!open {action.id}",
                  action=action).add()

        action.charges = charge

        await action.joueur.private_chan.send(
            f"Ton action {action.base.slug} vient d'être rechargée, "
            f"tu as maintenant {charge} charge(s) disponible(s) !"
        )

    config.session.commit()



class OpenClose(commands.Cog):
    """Commandes de gestion des votes et actions"""

    @commands.command()
    @tools.mjs_only
    async def open(self, ctx, qui, heure=None, heure_chain=None):
        """Lance un vote / des actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions commençant à ``heure``
                ``{id}``        pour une action spécifique
                ===========     ===========

            heure:
                - si ``qui == "cond"``, ``"maire"`` ou ``"loup"``,
                  programme en plus la fermeture à ``heure``
                  (et un rappel 30 minutes avant) ;
                - si ``qui == "action"``, il est obligatoire : heure des
                  actions à lancer (cf plus haut). Pour les actions, la
                  fermeture est de toute façon programmée le cas échéant
                  (``trigger_fin`` ``temporel`` ou ``delta``).

                Dans tous les cas, format ``HHh`` ou ``HHhMM``.

            heure_chain:
                permet de chaîner des votes : lance le vote immédiatement
                et programme sa fermeture à ``heure``, en appellant ``!close``
                de sorte à programmer une nouvelle ouverture le lendemain à
                ``heure_chain``, et ainsi de suite.
                Format ``HHh`` ou ``HHhMM``.

        Une sécurité empêche de lancer un vote ou une action déjà en cours.

        Cette commande a pour vocation première d'être exécutée
        automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas
        faire n'importe quoi (penser à envoyer / planifier la fermeture
        des votes, par exemple).

        Examples:
            - ``!open maire`` :        lance un vote maire maintenant
            - ``!open cond 19h`` :     lance un vote condamné maintenant
              et programme sa fermeture à 19h00 (ex. Juge Bègue)
            - ``!open cond 18h 10h`` : lance un vote condamné maintenant,
              programme sa fermeture à 18h00, et une prochaine ouverture
              à 10h qui se fermera à 18h, et ainsi de suite
            - ``!open action 19h`` :   lance toutes les actions
              commençant à 19h00
            - ``!open 122`` :          lance l'action d'ID 122

        """
        try:
            qui = Vote[qui.lower()]         # cond / maire / loups
        except KeyError:
            pass
        joueurs = await recup_joueurs("open", qui, heure)
        # Liste de joueurs (votes) ou dictionnaire joueur : action

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await tools.send_code_blocs(
            ctx,
            f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n"
            + str_joueurs
        )

        # Création utilisations & envoi messages
        for joueur in joueurs:
            chan = joueur.private_chan

            if isinstance(qui, Vote):
                action = joueur.action_vote(qui)
                if action.is_open:      # Sécurité : action ouverte depuis
                    continue
                util = Utilisation(action=action)
                util.add()
                util.open()

            if qui == Vote.cond:
                message = await chan.send(
                    f"{tools.montre()}  Le vote pour le condamné du "
                    f"jour est ouvert !  {config.Emoji.bucher} \n"
                    + (f"Tu as jusqu'à {heure} pour voter. \n"
                       if heure else "")
                    + tools.ital(f"Tape {tools.code('!vote (nom du joueur)')}"
                                 " ou utilise la réaction pour voter.")
                )
                await message.add_reaction(config.Emoji.bucher)

            elif qui == Vote.maire:
                message = await chan.send(
                    f"{tools.montre()}  Le vote pour l'élection du "
                    f"maire est ouvert !  {config.Emoji.maire} \n"
                    + (f"Tu as jusqu'à {heure} pour voter. \n"
                       if heure else "")
                    + tools.ital(
                        f"Tape {tools.code('!votemaire (nom du joueur)')} "
                        "ou utilise la réaction pour voter."
                    )
                )
                await message.add_reaction(config.Emoji.maire)

            elif qui == Vote.loups:
                message = await chan.send(
                    f"{tools.montre()}  Le vote pour la victime de "
                    f"cette nuit est ouvert !  {config.Emoji.lune} \n"
                    + (f"Tu as jusqu'à {heure} pour voter. \n"
                       if heure else "")
                    + tools.ital(
                        f"Tape {tools.code('!voteloups (nom du joueur)')} "
                        "ou utilise la réaction pour voter."
                    )
                )
                await message.add_reaction(config.Emoji.lune)

            else:       # Action
                for action in joueurs[joueur]:
                    await gestion_actions.open_action(action)

        config.session.commit()

        # Actions déclenchées par ouverture
        if isinstance(qui, Vote):
            for action in Action.query.filter(Action.base.has(
                    BaseAction.trigger_debut == ActionTrigger.open(qui))):
                await gestion_actions.open_action(action)

            for action in Action.query.filter(Action.base.has(
                    BaseAction.trigger_fin == ActionTrigger.open(qui))):
                await gestion_actions.close_action(action)

        # Réinitialise haros/candids
        items = []
        if qui == Vote.cond:
            items = CandidHaro.query.filter_by(
                type=CandidHaroType.haro).all()
        elif qui == Vote.maire:
            items = CandidHaro.query.filter_by(
                type=CandidHaroType.candidature).all()
        if items:
            CandidHaro.delete(*items)
            await tools.log(f"!open {qui.name} : haros/candids wiped")
            await config.Channel.haros.send(
                f"{config.Emoji.void}\n" * 30
                + "Nouveau vote, nouveaux haros !\n"
                + tools.ital(
                    "Les posts ci-dessus sont invalides pour le vote actuel. "
                    f"Utilisez {tools.code('!haro')} pour en relancer."
                )
            )

        # Programme fermeture
        if isinstance(qui, Vote) and heure:
            ts = tools.next_occurence(tools.heure_to_time(heure))
            Tache(timestamp=ts - datetime.timedelta(minutes=30),
                  commande=f"!remind {qui.name}").add()
            if heure_chain:
                Tache(timestamp=ts,
                      commande=f"!close {qui.name} {heure_chain} {heure}"
                ).add()
                # Programmera prochaine ouverture
            else:
                Tache(timestamp=ts, commande=f"!close {qui.name}").add()


    @commands.command()
    @tools.mjs_only
    async def close(self, ctx, qui, heure=None, heure_chain=None):
        """Ferme un vote / des actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions se terminant à ``heure``
                ``{id}``        pour une action spécifique
                ===========     ===========

            heure:
                - si ``qui == "cond"``, ``"maire"`` ou ``"loup"``,
                  programme en plus une prochaine ouverture à ``heure``
                  (et un rappel 30 minutes avant) ;
                - si ``qui == "action"``, il est obligatoire : heure des
                  actions à lancer (cf plus haut). Pour les actions, la
                  prochaine est de toute façon programmée le cas échéant
                  (cooldown à 0 et reste des charges).

                Dans tous les cas, format ``HHh`` ou ``HHhMM``.

            heure_chain:
                permet de chaîner des votes : ferme le vote immédiatement
                et programme une prochaine ouverture à ``heure``, en
                appellant ``!close`` de sorte à programmer une nouvelle
                fermeture le lendemain à ``heure_chain``, et ainsi de suite.
                Format ``HHh`` ou ``HHhMM``.

        Une sécurité empêche de fermer un vote ou une action
        qui n'est pas en cours.

        Cette commande a pour vocation première d'être exécutée
        automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas
        faire n'importe quoi (penser à envoyer / planifier la fermeture
        des votes, par exemple).

        Examples:
            - ``!close maire`` :        ferme le vote condamné maintenant
            - ``!close cond 10h`` :     ferme le vote condamné maintenant
              et programme une prochaine ouverture à 10h00
            - ``!close cond 10h 18h`` : ferme le vote condamné maintenant,
              programme une prochaine ouverture à 10h00, qui sera fermé à
              18h, puis une nouvelle ouverture à 10h, etc
            - ``!close action 22h`` :   ferme toutes les actions
              se terminant à 22h00
            - ``!close 122`` :          ferme l'action d'ID 122
        """
        try:
            qui = Vote[qui.lower()]         # cond / maire / loups
        except KeyError:
            pass
        joueurs = await recup_joueurs("close", qui, heure)

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await tools.send_code_blocs(
            ctx,
            f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n"
            + str_joueurs
        )

        # Fermeture utilisations et envoi messages
        for joueur in joueurs:
            chan = joueur.private_chan

            if isinstance(qui, Vote):
                action = joueur.action_vote(qui)
                if not action.is_open:      # Sécurité : action fermée depuis
                    continue
                util = joueur.action_vote(qui).utilisation_ouverte
                nom_cible = util.cible.nom if util.cible else "*non défini*"

                util.close()        # update direct pour empêcher de voter

            if qui == Vote.cond:
                await chan.send(
                    f"{tools.montre()}  Fin du vote pour le condamné du jour !"
                    f"\nVote définitif : {nom_cible}\n"
                    f"Les résultats arrivent dans l'heure !\n"
                )

            elif qui == Vote.maire:
                await chan.send(
                    f"{tools.montre()}  Fin du vote pour le maire ! \n"
                    f"Vote définitif : {nom_cible}"
                )

            elif qui == Vote.loups:
                await chan.send(
                    f"{tools.montre()}  Fin du vote pour la victime du soir !"
                    f"\nVote définitif : {nom_cible}"
                )

            else:       # Action
                for action in joueurs[joueur]:
                    await chan.send(
                        f"{tools.montre()}  Fin de la possiblité d'utiliser "
                        f"ton action {tools.code(action.base.slug)} ! \n"
                        f"Action définitive : {action.decision}"
                    )
                    await gestion_actions.close_action(action)

        config.session.commit()

        # Actions déclenchées par fermeture
        if isinstance(qui, Vote):
            for action in Action.query.filter(Action.base.has(
                  BaseAction.trigger_debut == ActionTrigger.close(qui))):
                await gestion_actions.open_action(action)

            for action in Action.query.filter(Action.base.has(
                  BaseAction.trigger_fin == ActionTrigger.close(qui))):
                await gestion_actions.close_action(action)

        # Programme prochaine ouverture
        if isinstance(qui, Vote) and heure:
            ts = tools.next_occurence(tools.heure_to_time(heure))
            if heure_chain:
                Tache(timestamp=ts,
                      commande=f"!open {qui.name} {heure_chain} {heure}").add()
                # Programmera prochaine fermeture
            else:
                Tache(timestamp=ts, commande=f"!open {qui.name}").add()


    @commands.command()
    @tools.mjs_only
    async def remind(self, ctx, qui, heure=None):
        """Envoi un rappel de vote / actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions se terminant à ``heure``
                ``{id}``        pour une action spécifique
                ===========     ===========

            heure: ne sert que dans le cas où <qui> == "action"
                (il est alors obligatoire), contrairement à !open et
                !close.
                Format HHh ou HHhMM.

        Le bot n'envoie un message qu'aux joueurs n'ayant pas encore
        voté / agi, si le vote ou l'action est bien en cours.

        Cette commande a pour vocation première d'être exécutée
        automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas
        faire n'importe quoi !.

        Examples:
            - ``!remind maire`` :      rappelle le vote maire maintenant
            - ``!remind action 22h`` : rappelle toutes les actions
              se terminant à 22h00
            - ``!remind 122`` :        rappelle l'action d'ID 122
        """
        try:
            qui = Vote[qui.lower()]         # cond / maire / loups
        except KeyError:
            pass
        joueurs = await recup_joueurs("remind", qui, heure)

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await ctx.send(tools.code_bloc(
            f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n"
            + str_joueurs
        ))

        for joueur in joueurs:
            chan = joueur.private_chan
            member = joueur.member

            if qui == Vote.cond:
                message = await chan.send(
                    f"⏰ {member.mention} Plus que 30 minutes pour voter "
                    "pour le condamné du jour ! 😱 \n"
                )
                await message.add_reaction(config.Emoji.bucher)

            elif qui == Vote.maire:
                message = await chan.send(
                    f"⏰ {member.mention} Plus que 30 minutes pour élire "
                    "le nouveau maire ! 😱 \n"
                )
                await message.add_reaction(config.Emoji.maire)

            elif qui == Vote.loups:
                message = await chan.send(
                    f"⏰ {member.mention} Plus que 30 minutes pour voter "
                    "pour la victime du soir ! 😱 \n"
                )
                await message.add_reaction(config.Emoji.lune)

            else:       # Action
                for action in joueurs[joueur]:
                    message = await chan.send(
                        f"⏰ {member.mention} Plus que 30 minutes pour "
                        f"utiliser ton action {tools.code(action.base.slug)}"
                        " ! 😱 \n"
                    )
                    await message.add_reaction(config.Emoji.action)


    @commands.command()
    @tools.mjs_only
    async def refill(self, ctx, motif, *, cible=None):
        """Recharger un/des pouvoirs rechargeables (COMMANDE BOT / MJ)

        Args:
            motif: ``"weekends"``, ``"forgeron"``, ``"rebouteux"``
                ou ``"divin"`` (forcer le refill car les MJs
                tout-puissants l'ont décidé)
            cible: ``"all"`` ou le nom d'un joueur
        """
        motif = motif.lower()

        if motif not in [*config.refills_full, *config.refills_one]:
            await ctx.send(f"{motif} n'est pas un motif valide")
            return

        if motif in config.refills_divins:
            if cible != "all":
                target = await tools.boucle_query_joueur(
                    ctx, cible=cible, message="Qui veux-tu recharger ?"
                )
                refillable = Action.query.filter(
                    Action.charges.isnot(None)).filter_by(joueur=target).all()
            else:
                m = await ctx.send(
                    "Tu as choisi de recharger le pouvoir de "
                    "TOUS les joueurs actifs, en es-tu sûr ?"
                )
                if await tools.yes_no(m):
                    refillable = Action.query.filter(
                        Action.charges.isnot(None)).all()

                else:
                    await ctx.send("Mission aborted.")
                    return

        else:       # refill WE, forgeron ou rebouteux
            if cible != "all":
                target = await tools.boucle_query_joueur(
                    ctx, cible=cible, message="Qui veux-tu recharger ?"
                )
                refillable = Action.query.filter(Action.base.has(
                    BaseAction.refill.contains(motif))).filter_by(
                    joueur=target).all()
            else:
                refillable = Action.query.filter(Action.base.has(
                    BaseAction.refill.contains(motif))).all()

        # do refill
        await tools.log(refillable, code=True,
                        prefixe=f"Refill {motif} {cible} :")

        await tools.send_code_blocs(
            ctx,
            "\n".join(f"- {action.base.slug}, id = {action.id} \n"
                      for action in refillable),
            prefixe="Action(s) répondant aux critères :\n"
        )

        await _do_refill(motif, refillable)



    @commands.command()
    @tools.mjs_only
    async def cparti(self, ctx):
        """Lance le jeu (COMMANDE MJ)

        - Programme les votes condamnés quotidiens (avec chaînage) 10h-18h
        - Programme un vote maire 10h-18h
        - Programme les actions au lancement du jeu (choix de mentor...)
          et permanentes (forgeron)... à 19h
        - Crée les "actions de vote", sans quoi !open plante

        À utiliser le jour du lancement après 10h (lance les premières
        actions le soir et les votes le lendemain)
        """

        message = await ctx.send(
            "C'est parti ?\n"
            "Les rôles ont bien été attribués et synchronisés ?"
            " (si non, le faire AVANT de valider)\n\n"
            "On est bien après 10h le jour du lancement ?\n\n"
            "Tu es conscient que tous les joueurs reçevront à 18h55 un message"
            " en mode « happy Hunger Games » ? (codé en dur parce que flemme)"
        )
        if not await tools.yes_no(message):
            await ctx.send("Mission aborted.")
            return

        message = await ctx.send(
            "Les actions des joueurs ont été attribuées à la synchronisation "
            "des rôles, mais les !open n'ont aucun impact tant que tout le "
            "monde est en `role_actif == False` sur le Tableau de bord.\n"
            "Il faut donc **passer tout le monde à `True` maintenant**"
            "(puis `!sync silent`) avant de continuer."
        )
        if not await tools.yes_no(message):
            await ctx.send("Mission aborted.")
            return

        taches = []
        r = "C'est parti !\n"

        n10 = tools.next_occurence(datetime.time(hour=10))
        n19 = tools.next_occurence(datetime.time(hour=19))

        # Programmation votes condamnés chainés 10h-18h
        r += "\nProgrammation des votes :\n"
        taches.append(Tache(timestamp=n10, commande="!open cond 18h 10h"))
        r += " - À 10h : !open cond 18h 10h\n"

        # Programmation votes loups chainés 19h-23h
        taches.append(Tache(timestamp=n19, commande="!open loups 23h 19h"))
        r += " - À 19h : !open loups 23h 19h\n"

        # Programmation premier vote maire 10h-17h
        taches.append(Tache(timestamp=n10, commande="!open maire 17h"))
        r += " - À 10h : !open maire 17h\n"

        # Programmation actions au lancement et actions permanentes
        r += "\nProgrammation des actions start / perma :\n"
        start_perma = Action.query.filter(
            Action.base.has(BaseAction.trigger_debut.in_(
                [ActionTrigger.start, ActionTrigger.perma]
        ))).all()
        for action in start_perma:
            r += (f" - À 19h : !open {action.id} "
                  f"(trigger_debut == {action.base.trigger_debut})\n")
            taches.append(Tache(timestamp=n19,
                                commande=f"!open {action.id}",
                                action=action))

        # Programmation refill weekends
        # r += "\nProgrammation des refills weekends :\n"
        # ts = tools.fin_pause() - datetime.timedelta(minutes=5)
        # taches.append(Tache(timestamp=ts,
        #                     commande=f"!refill weekends all"))
        # r += " - Dimanche à 18h55 : !refill weekends all\n"

        # Programmation envoi d'un message aux connards
        r += ("\nEt, à 18h50 : !send all [message de hype oue oue "
              "c'est génial]\n")
        taches.append(Tache(
            timestamp=(n19 - datetime.timedelta(minutes=10)),
            commande=(
                "!send all Ah {member.mention}... J'espère que tu "
                "es prêt(e), parce que la partie commence DANS 10 "
                " MINUTES !!! https://tenor.com/view/thehungergames-"
                "hungergames-thggifs-effie-gif-5114734"
            )
        ))
        await tools.log(r, code=True)

        # Drop (éventuel) et (re-)création actions de vote
        Action.query.filter_by(base=None).delete()
        actions = []
        for joueur in Joueur.query.all():
            for vote in Vote:
                actions.append(Action(joueur=joueur, vote=vote))

        Tache.add(*taches)      # On enregistre et programme le tout !
        Action.add(*actions)

        await ctx.send(
            f"C'est tout bon ! (détails dans {config.Channel.logs.mention})"
        )
